"""
Functions used to support easing
"""

from collections.abc import Callable
from dataclasses import dataclass
from math import cos, pi, sin

from .math import get_distance


@dataclass
class EasingData:
    """
    Data class for holding information about easing.
    """

    start_period: float
    cur_period: float
    end_period: float
    start_value: float
    end_value: float
    ease_function: Callable

    def reset(self) -> None:
        """
        Reset the easing data to its initial state.
        """
        self.cur_period = self.start_period


def linear(percent: float) -> float:
    """
    Function for linear easing.
    """
    return percent


def _flip(percent: float) -> float:
    return 1.0 - percent


def smoothstep(percent: float) -> float:
    """
    Function for smoothstep easing.
    """
    return percent**2 * (3.0 - 2.0 * percent)


def ease_in(percent: float) -> float:
    """
    Function for quadratic ease-in easing.
    """
    return percent**2


def ease_out(percent: float) -> float:
    """
    Function for quadratic ease-out easing.
    """
    return _flip(_flip(percent) * _flip(percent))


def ease_in_out(percent: float) -> float:
    """
    Function for quadratic easing in and out.
    """

    return 2 * percent**2 if percent < 0.5 else 1 - (-2 * percent + 2) ** 2 / 2


def ease_out_elastic(percent: float) -> float:
    """
    Function for elastic ease-out easing.
    """
    c4 = 2 * pi / 3
    result = 0.0
    if percent == 1:
        result = 1
    elif percent > 0:
        result = (2 ** (-10 * percent)) * sin((percent * 10 - 0.75) * c4) + 1
    return result


def ease_out_bounce(percent: float) -> float:
    """
    Function for a bouncy ease-out easing.
    """
    n1 = 7.5625
    d1 = 2.75

    if percent < 1 / d1:
        return n1 * percent * percent
    elif percent < 2 / d1:
        percent_modified = percent - 1.5 / d1
        return n1 * percent_modified * percent_modified + 0.75
    elif percent < 2.5 / d1:
        percent_modified = percent - 2.25 / d1
        return n1 * percent_modified * percent_modified + 0.9375
    else:
        percent_modified = percent - 2.625 / d1
        return n1 * percent_modified * percent_modified + 0.984375


def ease_in_back(percent: float) -> float:
    """
    Function for ease_in easing which moves back before moving forward.
    """
    c1 = 1.70158
    c3 = c1 + 1

    return c3 * percent * percent * percent - c1 * percent * percent


def ease_out_back(percent: float) -> float:
    """
    Function for ease_out easing which moves back before moving forward.
    """
    c1 = 1.70158
    c3 = c1 + 1

    return 1 + c3 * pow(percent - 1, 3) + c1 * pow(percent - 1, 2)


def ease_in_sin(percent: float) -> float:
    """
    Function for ease_in easing using a sin wave
    """
    return 1 - cos((percent * pi) / 2)


def ease_out_sin(percent: float) -> float:
    """
    Function for ease_out easing using a sin wave
    """
    return sin((percent * pi) / 2)


def ease_in_out_sin(percent: float) -> float:
    """
    Function for easing in and out using a sin wave
    """
    return -cos(percent * pi) * 0.5 + 0.5


def easing(percent: float, easing_data: EasingData) -> float:
    """
    Function for calculating return value for easing, given percent and easing data.
    """
    return easing_data.start_value + (
        easing_data.end_value - easing_data.start_value
    ) * easing_data.ease_function(percent)


def ease_angle(
    start_angle: float,
    end_angle: float,
    *,
    time=None,
    rate=None,
    ease_function: Callable = linear,
) -> EasingData | None:
    """
    Set up easing for angles.
    """
    while start_angle - end_angle > 180:
        end_angle += 360

    while start_angle - end_angle < -180:
        end_angle -= 360

    diff = abs(start_angle - end_angle)
    if diff == 0:
        return None

    if rate is not None:
        time = diff / rate

    if time is None:
        raise ValueError("Either the 'time' or the 'rate' parameter needs to be set.")

    easing_data = EasingData(
        start_value=start_angle,
        end_value=end_angle,
        start_period=0,
        cur_period=0,
        end_period=time,
        ease_function=ease_function,
    )
    return easing_data


def ease_angle_update(easing_data: EasingData, delta_time: float) -> tuple[bool, float]:
    """
    Update angle easing.
    """
    done = False
    easing_data.cur_period += delta_time
    easing_data.cur_period = min(easing_data.cur_period, easing_data.end_period)
    percent = easing_data.cur_period / easing_data.end_period

    angle = easing(percent, easing_data)

    if percent >= 1.0:
        done = True

        while angle > 360:
            angle -= 360

        while angle < 0:
            angle += 360

    return done, angle


def ease_value(
    start_value: float, end_value: float, *, time=None, rate=None, ease_function=linear
) -> EasingData:
    """
    Get an easing value
    """
    if rate is not None:
        diff = abs(start_value - end_value)
        time = diff / rate

    if time is None:
        raise ValueError("Either the 'time' or the 'rate' parameter needs to be set.")

    easing_data = EasingData(
        start_value=start_value,
        end_value=end_value,
        start_period=0,
        cur_period=0,
        end_period=time,
        ease_function=ease_function,
    )
    return easing_data


def ease_position(
    start_position, end_position, *, time=None, rate=None, ease_function=linear
) -> tuple[EasingData, EasingData]:
    """
    Get an easing position
    """
    distance = get_distance(start_position[0], start_position[1], end_position[0], end_position[1])

    if rate is not None:
        time = distance / rate

    easing_data_x = ease_value(
        start_position[0], end_position[0], time=time, ease_function=ease_function
    )
    easing_data_y = ease_value(
        start_position[1], end_position[1], time=time, ease_function=ease_function
    )

    return easing_data_x, easing_data_y


def ease_update(easing_data: EasingData, delta_time: float) -> tuple[bool, float]:
    """
    Update easing between two values/
    """
    easing_data.cur_period += delta_time
    easing_data.cur_period = min(easing_data.cur_period, easing_data.end_period)
    if easing_data.end_period == 0:
        percent = 1.0
        value = easing_data.end_value
    else:
        percent = easing_data.cur_period / easing_data.end_period
        value = easing(percent, easing_data)

    done = percent >= 1.0
    return done, value



"""
Sprite Collect Coins with Background

Simple program to show basic sprite usage.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_collect_coins_background
"""
import random
import arcade

PLAYER_SCALING = 0.75
COIN_SCALING = 0.4

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Collect Coins with Background Example"


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """ Initializer """

        # Call the parent class initializer
        super().__init__()

        # Background image will be stored in this variable
        self.background = arcade.load_texture(":resources:images/backgrounds/abstract_1.jpg")

        # Variables that will hold sprite lists
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=PLAYER_SCALING,
        )
        self.player_list = arcade.SpriteList()
        self.player_list.append(self.player_sprite)
        self.coin_list = arcade.SpriteList()

        # Set up the player info
        self.score = 0
        self.score_text = arcade.Text("Score: 0", 10, 20, arcade.color.WHITE, 14)

        # Don't show the mouse cursor
        self.window.set_mouse_visible(False)

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def reset(self):
        """Restart the game."""
        # Sprite lists
        self.coin_list.clear()

        # Set up the player
        self.score = 0
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 50

        for i in range(50):
            # Create the coin instance
            coin = arcade.Sprite(":resources:images/items/coinGold.png", scale=COIN_SCALING)

            # Position the coin
            coin.center_x = random.randrange(WINDOW_WIDTH)
            coin.center_y = random.randrange(WINDOW_HEIGHT)

            # Add the coin to the lists
            self.coin_list.append(coin)

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw the background texture
        arcade.draw_texture_rect(
            self.background,
            arcade.LBWH(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT),
        )

        # Draw all the sprites.
        self.coin_list.draw()
        self.player_list.draw()

        # Update the score text and draw it
        self.score_text.text = f"Score: {self.score}"
        self.score_text.draw()

    def on_mouse_motion(self, x, y, dx, dy):
        """
        Called whenever the mouse moves.
        """
        self.player_sprite.center_x = x
        self.player_sprite.center_y = y

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on the coin sprites (The sprites don't do much in this
        # example though.)
        self.coin_list.update()

        # Generate a list of all sprites that collided with the player.
        hit_list = arcade.check_for_collision_with_list(self.player_sprite, self.coin_list)

        # Loop through each colliding sprite, remove it, and add to the score.
        for coin in hit_list:
            coin.remove_from_sprite_lists()
            self.score += 1

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == arcade.key.R:
            self.reset()
        elif symbol == arcade.key.ESCAPE:
            self.window.close()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.reset()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Turn and Move Example.

Right-click to cause the tank to move to that point.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.turn_and_move
"""
import math
import arcade

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Turn and Move Example"

# Image might not be lined up right, set this to offset
IMAGE_ROTATION = -90


class Player(arcade.Sprite):
    """
    Sprite that turns and moves
    """
    def __init__(self):
        super().__init__(":resources:images/topdown_tanks/tank_green.png")

        # Destination point is where we are going
        self._destination_point = None

        # Max speed px / s
        self.speed = 300

        # what percent of the angle do we move by each frame
        self.rot_speed = 0.1

    @property
    def destination_point(self):
        return self._destination_point

    @destination_point.setter
    def destination_point(self, destination_point):
        self._destination_point = destination_point
        self.change_x = 0.0
        self.change_y = 0.0

    def update(self, delta_time: float = 1 / 60):
        """ Update the player """

        # If we have no destination, don't go anywhere.
        if not self._destination_point:
            self.change_x = 0
            self.change_y = 0
            return

        # Position the start at our current location
        start_x = self.center_x
        start_y = self.center_y

        # Get the destination location
        dest_x = self._destination_point[0]
        dest_y = self._destination_point[1]

        # Do math to calculate how to get the sprite to the destination.
        # Calculation the angle in radians between the start points
        # and end points. This is the angle the player will travel.
        target_angle = arcade.math.get_angle_degrees(start_x, start_y, dest_x, dest_y)
        current_angle = self.angle - IMAGE_ROTATION

        new_angle = arcade.math.lerp_angle(current_angle, target_angle, self.rot_speed)

        self.angle = new_angle + IMAGE_ROTATION
        angle_diff = abs(target_angle - new_angle)
        if  angle_diff < 0.1 or 359.9 < angle_diff:
            self.angle = target_angle + IMAGE_ROTATION
            target_radians = math.radians(target_angle)
            self.change_x = math.cos(-target_radians) * self.speed
            self.change_y = math.sin(-target_radians) * self.speed

        # Fine-tune our change_x/change_y if we are really close to destination
        # point and just need to set to that location.
        traveling = False
        if abs(self.center_x - dest_x) < abs(self.change_x * delta_time):
            self.center_x = dest_x
        else:
            self.center_x += self.change_x * delta_time
            traveling = True

        if abs(self.center_y - dest_y) < abs(self.change_y * delta_time):
            self.center_y = dest_y
        else:
            self.center_y += self.change_y * delta_time
            traveling = True

        # If we have arrived, then cancel our destination point
        if not traveling:
            self._destination_point = None


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.SAND

        self.player_sprite = None

        # Sprite Lists
        self.player_list = None

    def setup(self):
        """ Set up the game variables. Call to re-start the game. """

        # Sprite Lists
        self.player_list = arcade.SpriteList()
        self.player_sprite = Player()
        self.player_sprite.center_x = 300
        self.player_sprite.center_y = 300
        self.player_list.append(self.player_sprite)

    def on_draw(self):
        """
        Render the screen.
        """

        # This command should happen before we start drawing. It will clear
        # the screen to the background color, and erase what we drew last frame.
        self.clear()

        # Call draw() on all your sprite lists below
        self.player_list.draw()

    def on_update(self, delta_time):
        """
        All the logic to move, and the game logic goes here.
        """
        self.player_list.update(delta_time)

    def on_mouse_press(self, x, y, button, key_modifiers):
        """
        Called when the user presses a mouse button.
        """
        if button == arcade.MOUSE_BUTTON_RIGHT:
            self.player_sprite.destination_point = x, y


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()

if __name__ == "__main__":
    main()



"""
3D Example with offscreen rendering.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.3d_cube_with_cubes
"""

from pyglet.math import Mat4, Vec3

import arcade
from arcade.gl import geometry


class GameView(arcade.Window):

    def __init__(self, width, height, title):
        super().__init__(width, height, title, resizable=False)

        # Use the standard cube
        self.cube = geometry.cube()
        # Simple color lighting program for cube
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            uniform mat4 projection;
            uniform mat4 modelview;

            in vec3 in_position;
            in vec3 in_normal;
            in vec2 in_uv;

            out vec3 normal;
            out vec3 pos;
            out vec2 uv;

            void main() {
                vec4 p = modelview * vec4(in_position, 1.0);
                gl_Position = projection * p;
                mat3 m_normal = transpose(inverse(mat3(modelview)));
                normal = m_normal * in_normal;
                pos = p.xyz;
                uv = in_uv;
            }
            """,
            fragment_shader="""
            #version 330

            uniform sampler2D texture0;
            uniform bool use_texture;
            out vec4 fragColor;

            in vec3 normal;
            in vec3 pos;
            in vec2 uv;

            void main()
            {
                float l = dot(normalize(-pos), normalize(normal));
                if (use_texture) {
                    fragColor = vec4(texture(texture0, uv).rgb * (0.25 + abs(l) * 0.75), 1.0);
                } else {
                    fragColor = vec4(1.0) * (0.25 + abs(l) * 0.75);
                }
            }
            """,
        )
        # Program for drawing fullscreen quad with texture
        self.quad_program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_vert;
            in vec2 in_uv;
            out vec2 uv;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
                uv = in_uv;
            }
            """,
            fragment_shader="""
            #version 330

            uniform sampler2D texture0;
            in vec2 uv;
            out vec4 fragColor;

            void main() {
                fragColor = texture(texture0, uv);
            }
            """,
        )
        self.quad_fs = geometry.quad_2d_fs()

        self.on_resize(*self.get_size())
        self.frame = 0

        self.fbo1 = self.ctx.framebuffer(
            color_attachments=[self.ctx.texture(self.get_size())],
            depth_attachment=self.ctx.depth_texture(self.get_size()),
        )
        self.fbo2 = self.ctx.framebuffer(
            color_attachments=[self.ctx.texture(self.get_size())],
            depth_attachment=self.ctx.depth_texture(self.get_size()),
        )

    def on_draw(self):
        self.ctx.enable_only(self.ctx.CULL_FACE, self.ctx.DEPTH_TEST)

        # Draw the current cube using the last one as a texture
        self.fbo1.use()
        self.fbo1.clear(color_normalized=(1.0, 1.0, 1.0, 1.0))

        translate = Mat4.from_translation(Vec3(0, 0, -1.75))
        rx = Mat4.from_rotation(self.time, Vec3(1, 0, 0))
        ry = Mat4.from_rotation(self.time, Vec3(0, 1, 0))
        modelview = translate @ rx @ ry

        self.program["use_texture"] = 1
        self.fbo2.color_attachments[0].use()

        self.program["modelview"] = modelview
        self.cube.render(self.program)

        self.ctx.disable(self.ctx.DEPTH_TEST)

        # Draw the current cube texture
        self.use()
        self.clear()
        self.fbo1.color_attachments[0].use()
        self.quad_fs.render(self.quad_program)

        # Swap offscreen buffers
        self.fbo1, self.fbo2 = self.fbo2, self.fbo1
        self.frame += 1

    def on_resize(self, width, height):
        """Set up viewport and projection"""
        self.ctx.viewport = 0, 0, width, height
        self.program["projection"] = (
            Mat4.perspective_projection(self.aspect_ratio, 0.1, 100, fov=60)
        )


if __name__ == "__main__":
    GameView(720, 720, "3D Cube").run()



import re
from typing import TYPE_CHECKING, Iterable

from pyglet import gl

if TYPE_CHECKING:
    from .context import Context as ArcadeGlContext

from .exceptions import ShaderException
from .types import SHADER_TYPE_NAMES, PyGLenum


class ShaderSource:
    """
    GLSL source container for making source parsing simpler.

    We support locating out attributes, applying ``#defines`` values
    and injecting common source.

    .. note:::
        We do assume the source is neat enough to be parsed
        this way and don't contain several statements on one line.

    Args:
        ctx:
            The context this framebuffer belongs to
        source:
            The GLSL source code
        common:
            Common source code to inject
        source_type:
            The shader type
    """

    def __init__(
        self,
        ctx: "ArcadeGlContext",
        source: str,
        common: Iterable[str] | None,
        source_type: PyGLenum,
    ):
        self._source = source.strip()
        self._type = source_type
        self._lines = self._source.split("\n") if source else []
        self._out_attributes: list[str] = []

        if not self._lines:
            raise ValueError("Shader source is empty")

        self._version = self._find_glsl_version()

        # GLES specific modifications
        if ctx.gl_api == "gles":
            # TODO: Use the version from the context
            self._lines[0] = "#version 310 es"
            self._lines.insert(1, "precision mediump float;")

            if self._type == gl.GL_GEOMETRY_SHADER:
                self._lines.insert(1, "#extension GL_EXT_geometry_shader : require")

            if self._type == gl.GL_COMPUTE_SHADER:
                self._lines.insert(1, "precision mediump image2D;")

            self._version = self._find_glsl_version()

        # Inject common source
        self.inject_common_sources(common)

        if self._type in [gl.GL_VERTEX_SHADER, gl.GL_GEOMETRY_SHADER]:
            self._parse_out_attributes()

    @property
    def version(self) -> int:
        """The glsl version"""
        return self._version

    @property
    def out_attributes(self) -> list[str]:
        """The out attributes for this program"""
        return self._out_attributes

    def inject_common_sources(self, common: Iterable[str] | None) -> None:
        """
        Inject common source code into the shader source.

        Args:
            common:
                A list of common source code strings to inject
        """
        if not common:
            return

        # Find the main function
        for line_number, line in enumerate(self._lines):
            if "main()" in line:
                break
        else:
            raise ShaderException("No main() function found when injecting common source")

        # Insert all common sources
        for source in common:
            lines = source.split("\n")
            self._lines = self._lines[:line_number] + lines + self._lines[line_number:]

    def get_source(self, *, defines: dict[str, str] | None = None) -> str:
        """Return the shader source

        Args:
            defines: Defines to replace in the source.
        """
        if not defines:
            return "\n".join(self._lines)

        lines = ShaderSource.apply_defines(self._lines, defines)
        return "\n".join(lines)

    def _find_glsl_version(self) -> int:
        if self._lines[0].strip().startswith("#version"):
            try:
                return int(self._lines[0].split()[1])
            except Exception:
                pass

        source = "\n".join(f"{str(i + 1).zfill(3)}: {line} " for i, line in enumerate(self._lines))

        raise ShaderException(
            (
                "Cannot find #version in shader source. "
                "Please provide at least a #version 330 statement in the beginning of the shader.\n"
                f"---- [{SHADER_TYPE_NAMES[self._type]}] ---\n"
                f"{source}"
            )
        )

    @staticmethod
    def apply_defines(lines: list[str], defines: dict[str, str]) -> list[str]:
        """Locate and apply #define values

        Args:
            lines:
                List of source lines
            defines:
                dict with ``name: value`` pairs.
        """
        for nr, line in enumerate(lines):
            line = line.strip()
            if line.startswith("#define"):
                try:
                    name = line.split()[1]
                    value = defines.get(name, None)
                    if value is None:
                        continue

                    lines[nr] = "#define {} {}".format(name, str(value))
                except IndexError:
                    pass

        return lines

    def _parse_out_attributes(self):
        """
        Locates out attributes so we don't have to manually supply them.

        Note that this currently doesn't work for structs.
        """
        for line in self._lines:
            res = re.match(r"(layout(.+)\))?(\s+)?(out)(\s+)(\w+)(\s+)(\w+)", line.strip())
            if res:
                self._out_attributes.append(res.groups()[-1])



import math

from arcade import gl
from arcade.math import rotate_point
from arcade.types import RGBOrA255

from .helpers import _generic_draw_line_strip


def draw_arc_filled(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    color: RGBOrA255,
    start_angle: float,
    end_angle: float,
    tilt_angle: float = 0,
    num_segments: int = 128,
) -> None:
    """
    Draw a filled in arc. Useful for drawing pie-wedges, or Pac-Man.

    Args:
        center_x:
            x position that is the center of the arc.
        center_y:
            y position that is the center of the arc.
        width:
            width of the arc.
        height:
            height of the arc.
        color:
            A 3 or 4 length tuple of 0-255 channel values or a
            :py:class:`.Color` instance.
        start_angle:
            start angle of the arc in degrees.
        end_angle:
            end angle of the arc in degrees.
        tilt_angle:
            angle the arc is tilted (clockwise).
        num_segments:
            Number of line segments used to draw arc.
    """
    unrotated_point_list = [(0.0, 0.0)]

    start_segment = int(start_angle / 360 * num_segments)
    end_segment = int(end_angle / 360 * num_segments)

    for segment in range(start_segment, end_segment + 1):
        theta = 2.0 * 3.1415926 * segment / num_segments

        x = width * math.cos(theta) / 2
        y = height * math.sin(theta) / 2

        unrotated_point_list.append((x, y))

    if tilt_angle == 0:
        uncentered_point_list = unrotated_point_list
    else:
        uncentered_point_list = [
            rotate_point(point[0], point[1], 0, 0, tilt_angle) for point in unrotated_point_list
        ]

    point_list = [(point[0] + center_x, point[1] + center_y) for point in uncentered_point_list]

    _generic_draw_line_strip(point_list, color, gl.TRIANGLE_FAN)


def draw_arc_outline(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    color: RGBOrA255,
    start_angle: float,
    end_angle: float,
    border_width: float = 1,
    tilt_angle: float = 0,
    num_segments: int = 128,
) -> None:
    """
    Draw the outside edge of an arc. Useful for drawing curved lines.

    Args:
        center_x:
            x position that is the center of the arc.
        center_y:
            y position that is the center of the arc.
        width:
            width of the arc.
        height:
            height of the arc.
        color:
            A 3 or 4 length tuple of 0-255 channel values or a
            :py:class:`.Color` instance.
        start_angle:
            Start angle of the arc in degrees.
        end_angle:
            End angle of the arc in degrees.
        border_width:
            width of line in pixels.
        tilt_angle:
            Angle the arc is tilted (clockwise).
        num_segments:
            Number of triangle segments that make up this circle.
            Higher is better quality, but slower render time.
    """
    unrotated_point_list = []

    start_segment = int(start_angle / 360 * num_segments)
    end_segment = int(end_angle / 360 * num_segments)

    inside_width = (width - border_width / 2) / 2
    outside_width = (width + border_width / 2) / 2
    inside_height = (height - border_width / 2) / 2
    outside_height = (height + border_width / 2) / 2

    for segment in range(start_segment, end_segment + 1):
        theta = 2.0 * math.pi * segment / num_segments

        x1 = inside_width * math.cos(theta)
        y1 = inside_height * math.sin(theta)

        x2 = outside_width * math.cos(theta)
        y2 = outside_height * math.sin(theta)

        unrotated_point_list.append((x1, y1))
        unrotated_point_list.append((x2, y2))

    if tilt_angle == 0:
        uncentered_point_list = unrotated_point_list
    else:
        uncentered_point_list = [
            rotate_point(point[0], point[1], 0, 0, tilt_angle) for point in unrotated_point_list
        ]

    point_list = [(point[0] + center_x, point[1] + center_y) for point in uncentered_point_list]

    _generic_draw_line_strip(point_list, color, gl.TRIANGLE_STRIP)



import array

from arcade import gl
from arcade.color import WHITE
from arcade.math import rotate_point
from arcade.sprite import BasicSprite
from arcade.texture import Texture
from arcade.texture_atlas.base import TextureAtlasBase
from arcade.types import LBWH, LRBT, XYWH, Color, Point2List, Rect, RGBOrA255
from arcade.window_commands import get_window

from .helpers import _generic_draw_line_strip


def draw_texture_rect(
    texture: Texture,
    rect: Rect,
    *,
    color: Color = WHITE,
    angle=0.0,
    blend=True,
    alpha=255,
    pixelated=False,
    atlas: TextureAtlasBase | None = None,
) -> None:
    """
    Draw a texture on a rectangle.

    Args:
        texture:
            The texture to draw.
        rect:
            Rectangle to draw the texture on.
        color:
             Color multiplier for the texture. Defaults to white.
        angle:
            Rotation of the texture in degrees. Defaults to zero.
        blend:
            If True, enable alpha blending. Defaults to True.
        alpha:
            Transparency of image. 0 is fully transparent, 255 (default) is fully opaque.
        atlas:
            The texture atlas the texture resides in.
            if not supplied the default texture atlas is used
    """
    ctx = get_window().ctx

    # Clamp alpha to 0-255
    alpha_normalized = max(0, min(255, alpha)) / 255.0

    if blend:
        ctx.enable(ctx.BLEND)
    else:
        ctx.disable(ctx.BLEND)

    atlas = atlas or ctx.default_atlas
    program = ctx.sprite_program_single

    texture_id, _ = atlas.add(texture)
    if pixelated:
        atlas.texture.filter = gl.NEAREST, gl.NEAREST
        program.set_uniform_safe("uv_offset_bias", 0.0)
    else:
        atlas.texture.filter = gl.LINEAR, gl.LINEAR
        program.set_uniform_safe("uv_offset_bias", 1.0)

    atlas.texture.use(unit=0)
    atlas.use_uv_texture(unit=1)

    geometry = ctx.geometry_empty
    program["pos"] = rect.x, rect.y, 0
    program["color"] = color.normalized
    program["size"] = rect.width, rect.height
    program["angle"] = angle
    program["texture_id"] = float(texture_id)
    program["spritelist_color"] = 1.0, 1.0, 1.0, alpha_normalized

    geometry.render(program, mode=gl.POINTS, vertices=1)

    if blend:
        ctx.disable(ctx.BLEND)


def draw_sprite(
    sprite: BasicSprite,
    *,
    blend: bool = True,
    alpha=255,
    pixelated=False,
    atlas: TextureAtlasBase | None = None,
) -> None:
    """
    Draw a sprite.

    Args:
        sprite:
            The sprite to draw.
        blend:
            Draw the sprite with or without alpha blending
        alpha:
            Fade the sprite from completely transparent to opaque (range: 0 to 255)
        pixelated:
            If true the sprite will be render in pixelated style. Otherwise smooth/linear
        atlas:
            The texture atlas the texture resides in.
            if not supplied the default texture atlas is used
    """
    draw_texture_rect(
        sprite.texture,
        rect=XYWH(sprite.center_x, sprite.center_y, sprite.width, sprite.height),
        color=sprite.color,
        angle=sprite._angle,
        blend=blend,
        alpha=alpha,
        pixelated=pixelated,
        atlas=atlas,
    )


def draw_sprite_rect(
    sprite: BasicSprite,
    rect: Rect,
    *,
    blend: bool = True,
    alpha=255,
    pixelated=False,
    atlas: TextureAtlasBase | None = None,
) -> None:
    """Draw a sprite.

    Args:
        sprite:
            The sprite to draw.
        rect:
            The location and size of the sprite
        blend:
            Draw the sprite with or without alpha blending
        alpha:
            Fade the sprite from completely transparent to opaque (range: 0 to 255)
        pixelated:
            If true the sprite will be render in pixelated style.
            Otherwise smooth/linear
        atlas:
            The texture atlas the texture resides in.
            if not supplied the default texture atlas is used
    """
    draw_texture_rect(
        sprite.texture,
        rect=rect,
        color=sprite.color,
        angle=sprite._angle,
        blend=blend,
        alpha=alpha,
        pixelated=pixelated,
        atlas=atlas,
    )


def draw_lrbt_rectangle_outline(
    left: float,
    right: float,
    bottom: float,
    top: float,
    color: RGBOrA255,
    border_width: float = 1,
) -> None:
    """
    Draw a rectangle by specifying left, right, bottom and top edges.

    Args:
        left:
            The x coordinate of the left edge of the rectangle.
        right:
            The x coordinate of the right edge of the rectangle.
        bottom:
            The y coordinate of the rectangle bottom.
        top:
            The y coordinate of the top of the rectangle.
        color:
            The outline color as an RGBA :py:class:`tuple`, RGB
            :py:class:`tuple`, or a :py:class:`.Color` instance.
        border_width:
            The width of the border in pixels. Defaults to one.
    Raises:
        ValueError: Raised if left > right or top < bottom.
    """
    if left > right:
        raise ValueError("Left coordinate must be less than or equal to the right coordinate")

    if bottom > top:
        raise ValueError("Bottom coordinate must be less than or equal to the top coordinate")

    draw_rect_outline(LRBT(left, right, bottom, top), color, border_width)


def draw_lbwh_rectangle_outline(
    left: float,
    bottom: float,
    width: float,
    height: float,
    color: RGBOrA255,
    border_width: float = 1,
) -> None:
    """
    Draw a rectangle extending from bottom left to top right

    Args:
        left:
            The x coordinate of the left edge of the rectangle.
        bottom:
            The y coordinate of the bottom of the rectangle.
        width:
            The width of the rectangle.
        height:
            The height of the rectangle.
        color:
            The outline color as an RGBA :py:class:`tuple`, RGB
            :py:class:`tuple`, or a :py:class:`.Color` instance.
        border_width:
            The width of the border in pixels. Defaults to one.
    """
    draw_rect_outline(LBWH(left, bottom, width, height), color, border_width)


def draw_lrbt_rectangle_filled(
    left: float, right: float, bottom: float, top: float, color: RGBOrA255
) -> None:
    """
    Draw a rectangle by specifying left, right, bottom and top edges.

    Args:
        left:
            The x coordinate of the left edge of the rectangle.
        right:
            The x coordinate of the right edge of the rectangle.
        bottom:
            The y coordinate of the rectangle bottom.
        top:
            The y coordinate of the top of the rectangle.
        color:
            The fill color as an RGBA :py:class:`tuple`,
            RGB :py:class:`tuple`, or a :py:class:`.Color` instance.
    Raises:
        ValueError: Raised if left > right or top < bottom.
    """
    if left > right:
        raise ValueError(
            f"Left coordinate {left} must be less than or equal to the right coordinate {right}"
        )

    if bottom > top:
        raise ValueError(
            f"Bottom coordinate {bottom} must be less than or equal to the top coordinate {top}"
        )

    draw_rect_filled(LRBT(left, right, bottom, top), color)


def draw_lbwh_rectangle_filled(
    left: float, bottom: float, width: float, height: float, color: RGBOrA255
) -> None:
    """
    Draw a filled rectangle extending from bottom left to top right

    Args:
        left:
            The x coordinate of the left edge of the rectangle.
        bottom:
            The y coordinate of the bottom of the rectangle.
        width:
            The width of the rectangle.
        height:
            The height of the rectangle.
        color:
            The fill color as an RGBA :py:class:`tuple`, RGB
            :py:class:`tuple`, :py:class:`.Color` instance
    """
    draw_rect_filled(LBWH(left, bottom, width, height), color)


def draw_rect_outline(
    rect: Rect, color: RGBOrA255, border_width: float = 1, tilt_angle: float = 0
) -> None:
    """
    Draw a rectangle outline.

    Args:
        rect:
            The rectangle to draw. a :py:class:`~arcade.Rect` instance.
        color:
            The outline color as an RGBA :py:class:`tuple`,
            RGB :py:class:`tuple`, or :py:class:`.Color` instance.
        border_width:
            width of the lines, in pixels.
        tilt_angle:
            rotation of the rectangle. Defaults to zero (clockwise).
    """

    HALF_BORDER = border_width / 2
    # Extremely unrolled code below

    # fmt: off
    left   = rect.left
    right  = rect.right
    bottom = rect.bottom
    top    = rect.top
    x      = rect.x
    y      = rect.y

    # o = outer, i = inner
    #
    # o_left, o_top                  o_right, o_top
    #  +----------------------------------------+
    #  | i_left, i_top          i_right, i_top  |
    #  |  +----------------------------------+  |
    #  |  |                                  |  |
    #  |  |                                  |  |
    #  |  |                                  |  |
    #  |  +----------------------------------+  |
    #  | i_left, i_bottom    i_right , i_bottom |
    #  +----------------------------------------+
    # o_left, o_bottom               o_right, o_bottom

    i_left   = left   + HALF_BORDER
    i_bottom = bottom + HALF_BORDER
    i_right  = right  - HALF_BORDER
    i_top    = top    - HALF_BORDER

    o_left   = left   - HALF_BORDER
    o_bottom = bottom - HALF_BORDER
    o_right  = right  + HALF_BORDER
    o_top    = top    + HALF_BORDER

    # Declared separately because the code below seems to break mypy
    point_list: Point2List

    # This is intentionally unrolled to minimize repacking tuples
    if tilt_angle == 0:
        point_list = (
            (o_left  , o_top   ),
            (i_left  , i_top   ),
            (o_right , o_top   ),
            (i_right , i_top   ),
            (o_right , o_bottom),
            (i_right , i_bottom),
            (o_left  , o_bottom),
            (i_left  , i_bottom),
            (o_left  , o_top   ),
            (i_left  , i_top   )
        )
    else:
        point_list = (
            rotate_point(o_left   , o_top   , x, y, tilt_angle),
            rotate_point(i_left   , i_top   , x, y, tilt_angle),
            rotate_point(o_right  , o_top   , x, y, tilt_angle),
            rotate_point(i_right  , i_top   , x, y, tilt_angle),
            rotate_point(o_right  , o_bottom, x, y, tilt_angle),
            rotate_point(i_right  , i_bottom, x, y, tilt_angle),
            rotate_point(o_left   , o_bottom, x, y, tilt_angle),
            rotate_point(i_left   , i_bottom, x, y, tilt_angle),
            rotate_point(o_left   , o_top   , x, y, tilt_angle),
            rotate_point(i_left   , i_top   , x, y, tilt_angle)
        )
    # fmt: on
    _generic_draw_line_strip(point_list, color, gl.TRIANGLE_STRIP)


def draw_rect_filled(rect: Rect, color: RGBOrA255, tilt_angle: float = 0) -> None:
    """
    Draw a filled-in rectangle.

    Args:
        rect:
            The rectangle to draw. a :py:class:`~arcade.Rect` instance.
        color:
            The fill color as an RGBA :py:class:`tuple`,
            RGB :py:class:`tuple`, or :py:class:`.Color` instance.
        tilt_angle:
            rotation of the rectangle (clockwise). Defaults to zero.
    """
    # Fail if we don't have a window, context, or right GL abstractions
    window = get_window()
    ctx = window.ctx
    program = ctx.shape_rectangle_filled_unbuffered_program  # type: ignore
    geometry = ctx.shape_rectangle_filled_unbuffered_geometry
    buffer = ctx.shape_rectangle_filled_unbuffered_buffer  # type: ignore

    # Validate & normalize to a pass the shader an RGBA float uniform
    color_normalized = Color.from_iterable(color).normalized

    ctx.enable(ctx.BLEND)

    # Pass data to the shader
    program["color"] = color_normalized
    program["shape"] = rect.width, rect.height, tilt_angle
    buffer.orphan()
    buffer.write(data=array.array("f", (rect.x, rect.y)))

    geometry.render(program, mode=ctx.POINTS, vertices=1)

    ctx.disable(ctx.BLEND)


# These might be "oddly specific" and also needs docstrings. Disabling or 3.0.0

# def draw_rect_outline_kwargs(
#     color: RGBAOrA255 = WHITE, border_width: int = 1, tilt_angle: float = 0, **kwargs: AsFloat
# ) -> None:
#     rect = Rect.from_kwargs(**kwargs)
#     draw_rect_outline(rect, color, border_width, tilt_angle)


# def draw_rect_filled_kwargs(
#     color: RGBAOrA255 = WHITE, tilt_angle: float = 0, **kwargs: AsFloat
# ) -> None:
#     rect = Rect.from_kwargs(**kwargs)
#     draw_rect_filled(rect, color, tilt_angle)



from pyglet.math import Vec2

from arcade.experimental import Shadertoy


class CRTFilter:
    """
    CRT Filter

    Args:
        width: Width of the screen
        height: Height of the screen
        resolution_down_scale:
        hard_scan: -8.0 soft, -16.0 medium
        hard_pix: Hardness of pixels in the scan line. -2.0 soft, -4.0 medium
        display_warp: Display warp. 0 = None, 1.0/8.0 = extreme
    """

    def __init__(
        self,
        width: int,
        height: int,
        resolution_down_scale: float = 6.0,
        hard_scan: float = -8.0,
        hard_pix: float = -3.0,
        display_warp: Vec2 = Vec2(1.0 / 32.0, 1.0 / 24.0),
        mask_dark: float = 0.5,
        mask_light: float = 1.5,
    ):
        self.shadertoy = Shadertoy.create_from_file(
            (width, height), ":resources:shaders/shadertoy/crt_monitor_filter.glsl"
        )
        self.fbo = self.shadertoy.ctx.framebuffer(
            color_attachments=[self.shadertoy.ctx.texture((width, height), components=4)]
        )
        self.shadertoy.channel_0 = self.fbo.color_attachments[0]

        self.shadertoy.program["resolutionDownScale"] = resolution_down_scale
        self.shadertoy.program["hardScan"] = hard_scan
        self.shadertoy.program["hardPix"] = hard_pix
        self.shadertoy.program["warp"] = display_warp
        self.shadertoy.program["maskDark"] = mask_dark
        self.shadertoy.program["maskLight"] = mask_light

    def use(self):
        self.fbo.use()

    def clear(self):
        self.fbo.clear()

    def draw(self, time: float = 0, target=None):
        self.shadertoy.render()



class PyMunk:
    """Object used to hold pymunk info for a sprite."""

    __slots__ = (
        "damping",
        "gravity",
        "max_velocity",
        "max_horizontal_velocity",
        "max_vertical_velocity",
    )

    def __init__(self):
        self.damping: float | None = None
        """
        Natural damping.

        Damping is applied to objects to simulate the natural deceleration of
        objects due to friction.
        """

        self.gravity: tuple[float, float] | None = None
        """Gravity applied to the object."""

        self.max_velocity: float | None = None
        """Maximum velocity allowed."""

        self.max_horizontal_velocity: float | None = None
        """Maximum horizontal velocity allowed."""

        self.max_vertical_velocity: float | None = None
        """Maximum vertical velocity allowed."""


class PymunkMixin:
    """A mixin class that adds Pymunk physics to a sprite."""

    def __init__(self) -> None:
        self.pymunk = PyMunk()
        """Object used to hold pymunk info for a sprite."""
        self.force = [0.0, 0.0]
        """force vector used by pymunk"""

    def pymunk_moved(self, physics_engine, dx: float, dy: float, d_angle: float) -> None:
        """
        Called by the pymunk physics engine if this sprite moves.

        Args:
            physics_engine (PymunkPhysicsEngine): The physics engine that is calling this method.
            dx (float): The change in x position.
            dy (float): The change in y position.
            d_angle (float): The change in angle.
        """
        pass



"""
Platformer Game

python -m arcade.examples.platform_tutorial.02_draw_sprites
"""
import arcade

# Constants
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Platformer"


class GameView(arcade.Window):
    """
    Main application class.
    """

    def __init__(self):

        # Call the parent class and set up the window
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

        # Variable to hold our texture for our player
        self.player_texture = arcade.load_texture(
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_idle.png"
        )

        # Separate variable that holds the player sprite
        self.player_sprite = arcade.Sprite(self.player_texture)
        self.player_sprite.center_x = 64
        self.player_sprite.center_y = 128

        self.background_color = arcade.csscolor.CORNFLOWER_BLUE

    def setup(self):
        """Set up the game here. Call this function to restart the game."""
        pass

    def on_draw(self):
        """Render the screen."""

        # Clear the screen to the background color
        self.clear()

        # Draw our sprites
        arcade.draw_sprite(self.player_sprite)


def main():
    """Main function"""
    window = GameView()
    window.setup()
    arcade.run()


if __name__ == "__main__":
    main()



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, TypeVar, overload

from arcade.gui.property import DictProperty
from arcade.gui.widgets import UIWidget


@dataclass
class UIStyleBase:
    """Base class for styles to ensure a general interface and implement additional magic.

    Support dict like access syntax.

    A styled widget should own a dataclass, which subclasses this class
    """

    @overload
    def get(self, key, default: str) -> str: ...

    @overload
    def get(self, key, default: Any) -> Any: ...

    def get(self, key, default=None):
        """Get a value from the style, if not available return default."""
        return self[key] if key in self else default

    def __contains__(self, item):
        return hasattr(self, item)

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        setattr(self, key, value)


StyleRef = TypeVar("StyleRef", bound=UIStyleBase)


class UIStyledWidget(UIWidget, Generic[StyleRef]):
    """Baseclass for widgets to be styled.

    A styled widget should own a dataclass, which subclasses UIStyleBase.
    The style dict should contain the different states of the widget.

    The widget should implement py:method::`get_current_state()`,
    which should return the current state of the widget.

    Args:
        style: A mapping of states to styles
        **kwargs: passed to UIWidget

    """

    # TODO detect StyleBase changes, so that style changes can trigger rendering.
    style = DictProperty[str, StyleRef]()

    def __init__(self, *, style: dict[str, StyleRef], **kwargs):
        self.style = style
        super().__init__(**kwargs)

    @abstractmethod
    def get_current_state(self) -> str:
        """
        Return the current state of the widget. These should be contained in the style dict.

        Well known states:
        - normal
        - hover
        - press
        - disabled
        """
        pass

    def get_current_style(self) -> StyleRef | None:
        """Return style based on any state of the widget"""
        return self.style.get(self.get_current_state(), None)



"""
Line of Sight

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.line_of_sight
"""

import arcade
import random

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Line of Sight"

MOVEMENT_SPEED = 5

VIEWPORT_MARGIN = 250
HORIZONTAL_BOUNDARY = WINDOW_WIDTH / 2.0 - VIEWPORT_MARGIN
VERTICAL_BOUNDARY = WINDOW_HEIGHT / 2.0 - VIEWPORT_MARGIN
# If the player moves further than this boundary away from
# the camera we use a constraint to move the camera
CAMERA_BOUNDARY = arcade.LRBT(
    -HORIZONTAL_BOUNDARY,
    HORIZONTAL_BOUNDARY,
    -VERTICAL_BOUNDARY,
    VERTICAL_BOUNDARY,
)


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Initializer
        """

        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = None
        self.wall_list = None
        self.enemy_list = None

        # Set up the player info
        self.player = None

        # Track the current state of what key is pressed
        self.left_pressed = False
        self.right_pressed = False
        self.up_pressed = False
        self.down_pressed = False

        self.physics_engine = None

        # Camera for scrolling
        self.camera = None

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Camera
        self.camera = arcade.Camera2D()

        # Sprite lists
        self.player_list = arcade.SpriteList()
        self.wall_list = arcade.SpriteList(use_spatial_hash=True)
        self.enemy_list = arcade.SpriteList()

        # Set up the player
        self.player = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING,
        )
        self.player.center_x = 50
        self.player.center_y = 350
        self.player_list.append(self.player)

        # Set enemies
        enemy = arcade.Sprite(
            ":resources:images/animated_characters/zombie/zombie_idle.png",
            scale=SPRITE_SCALING,
        )
        enemy.center_x = 350
        enemy.center_y = 350
        self.enemy_list.append(enemy)

        spacing = 200
        for column in range(10):
            for row in range(10):
                sprite = arcade.Sprite(
                    ":resources:images/tiles/grassCenter.png",
                    scale=SPRITE_SCALING,
                )

                x = (column + 1) * spacing
                y = (row + 1) * sprite.height

                sprite.center_x = x
                sprite.center_y = y
                if random.randrange(100) > 20:
                    self.wall_list.append(sprite)

        self.physics_engine = arcade.PhysicsEngineSimple(
            self.player,
            self.wall_list,
        )

    def on_draw(self):
        """
        Render the screen.
        """
        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.player_list.draw()
        self.wall_list.draw()
        self.enemy_list.draw()

        for enemy in self.enemy_list:
            if arcade.has_line_of_sight(
                self.player.position, enemy.position, self.wall_list
            ):
                color = arcade.color.RED
            else:
                color = arcade.color.WHITE
            arcade.draw_line(self.player.center_x,
                             self.player.center_y,
                             enemy.center_x,
                             enemy.center_y,
                             color,
                             2)

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Calculate speed based on the keys pressed
        self.player.change_x = 0
        self.player.change_y = 0

        if self.up_pressed and not self.down_pressed:
            self.player.change_y = MOVEMENT_SPEED
        elif self.down_pressed and not self.up_pressed:
            self.player.change_y = -MOVEMENT_SPEED
        if self.left_pressed and not self.right_pressed:
            self.player.change_x = -MOVEMENT_SPEED
        elif self.right_pressed and not self.left_pressed:
            self.player.change_x = MOVEMENT_SPEED

        self.physics_engine.update()

        # --- Manage Scrolling ---
        self.camera.position = arcade.camera.grips.constrain_boundary_xy(
            self.camera.view_data, CAMERA_BOUNDARY, self.player.position
        )
        self.camera.use()

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP:
            self.up_pressed = True
        elif key == arcade.key.DOWN:
            self.down_pressed = True
        elif key == arcade.key.LEFT:
            self.left_pressed = True
        elif key == arcade.key.RIGHT:
            self.right_pressed = True

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP:
            self.up_pressed = False
        elif key == arcade.key.DOWN:
            self.down_pressed = False
        elif key == arcade.key.LEFT:
            self.left_pressed = False
        elif key == arcade.key.RIGHT:
            self.right_pressed = False


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Indexed geometry example.

Example using index buffers with 1, 2, and 4 byte size.
In most cases we can get away with unsigned 32 bit integers.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.geometry_index_buffers
"""

from array import array

import arcade
from arcade.gl import BufferDescription


class IndexBufferExample(arcade.Window):

    def __init__(self, width, height, title):
        super().__init__(width, height, title)
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_position;
            in vec3 in_color;

            out vec3 v_color;

            void main() {
                gl_Position = vec4(in_position, 0.0, 1.0);
                v_color = in_color;
            }
            """,
            fragment_shader="""
            #version 330

            out vec4 out_color;
            in vec3 v_color;

            void main() {
                out_color = vec4(v_color, 1.0);
            }
            """,
        )

        # Vertices
        # fmt: off
        self.vertex_buffer = self.ctx.buffer(data=array(
            'f',
            [
                # x  y    r    g    b
                -1.0, -1.0, 1.0, 0.0, 0.0,  # lower left
                1, -1.0, 0.0, 1.0, 0.0,  # lower right
                1.0, 1.0, 0.0, 0.0, 1.0,  # upper right
                -1.0, 1.0, 0.0, 1.0, 1.0,  # upper left
                0, 0, 0, 0, 0,  # dummy data for testing
            ]
        ))
        # fmt: on

        self.ibo_8 = self.ctx.buffer(data=array("B", [3, 0, 2, 1]))
        self.ibo_16 = self.ctx.buffer(data=array("H", [3, 0, 2, 1]))
        self.ibo_32 = self.ctx.buffer(data=array("I", [3, 0, 2, 1]))

        self.vao_32 = self.ctx.geometry(
            [
                BufferDescription(self.vertex_buffer, "2f 3f", ["in_position", "in_color"]),
            ],
            index_buffer=self.ibo_32,
            index_element_size=4,
            mode=self.ctx.TRIANGLE_STRIP,
        )
        self.vao_16 = self.ctx.geometry(
            [
                BufferDescription(self.vertex_buffer, "2f 3f", ["in_position", "in_color"]),
            ],
            index_buffer=self.ibo_16,
            index_element_size=2,
            mode=self.ctx.TRIANGLE_STRIP,
        )
        self.vao_8 = self.ctx.geometry(
            [
                BufferDescription(self.vertex_buffer, "2f 3f", ["in_position", "in_color"]),
            ],
            index_buffer=self.ibo_16,
            index_element_size=2,
            mode=self.ctx.TRIANGLE_STRIP,
        )
        self.frames = 0

    def on_draw(self):
        self.clear()

        mode = self.frames % 3
        if mode == 0:
            self.vao_8.render(self.program, mode=self.ctx.TRIANGLE_STRIP)
        elif mode == 1:
            self.vao_16.render(self.program, mode=self.ctx.TRIANGLE_STRIP)
        elif mode == 2:
            self.vao_32.render(self.program, mode=self.ctx.TRIANGLE_STRIP)

        self.frames += 1


if __name__ == "__main__":
    IndexBufferExample(1280, 720, "Index Buffers: 8, 16 and 32 bit")
    arcade.run()



"""
Show a timer on-screen.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.timer
"""
import arcade
from arcade.clock import GLOBAL_CLOCK

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Timer Example"


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        super().__init__()
        # What time to start the timer
        self.start_time: float = 0.0
        self.timer_text = arcade.Text(
            text="00:00:00",
            x=WINDOW_WIDTH // 2,
            y=WINDOW_HEIGHT // 2 - 50,
            color=arcade.color.WHITE,
            font_size=100,
            anchor_x="center",
        )
        self.background_color = arcade.color.ALABAMA_CRIMSON

    def reset(self):
        self.start_time = GLOBAL_CLOCK.time

    def on_draw(self):
        """ Use this function to draw everything to the screen. """
        # Clear all pixels in the window
        self.clear()

        # Draw the timer text
        self.timer_text.draw()

    def on_update(self, delta_time):
        """
        All the logic to move, and the game logic goes here.
        """
        # Accumulate the total time
        elapsed = GLOBAL_CLOCK.time_since(self.start_time)

        # Calculate minutes
        minutes = int(elapsed) // 60

        # Calculate seconds by using a modulus (remainder)
        seconds = int(elapsed) % 60

        # Calculate 100ths of a second
        seconds_100s = int((elapsed - seconds) * 100)

        # Use string formatting to create a new text string for our timer
        self.timer_text.text = f"{minutes:02d}:{seconds:02d}:{seconds_100s:02d}"

def main():
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    game = GameView()
    game.reset()

    window.show_view(game)
    arcade.run()


if __name__ == "__main__":
    main()



"""
The main window class that all object-oriented applications should
derive from.
"""

from __future__ import annotations

import logging
import os
import time
from collections.abc import Sequence
from typing import TYPE_CHECKING

import pyglet
import pyglet.gl as gl
import pyglet.window.mouse
from pyglet.display.base import Screen, ScreenMode
from pyglet.event import EVENT_HANDLE_STATE, EVENT_UNHANDLED
from pyglet.window import MouseCursor

import arcade
from arcade.clock import GLOBAL_CLOCK, GLOBAL_FIXED_CLOCK, _setup_clock, _setup_fixed_clock
from arcade.color import BLACK
from arcade.context import ArcadeContext
from arcade.types import LBWH, Color, Rect, RGBANormalized, RGBOrA255
from arcade.utils import is_raspberry_pi
from arcade.window_commands import get_display_size, set_window

if TYPE_CHECKING:
    from arcade.camera import Projector
    from arcade.camera.default import DefaultProjector
    from arcade.start_finish_data import StartFinishRenderData

LOG = logging.getLogger(__name__)

MOUSE_BUTTON_LEFT = 1
MOUSE_BUTTON_MIDDLE = 2
MOUSE_BUTTON_RIGHT = 4

_window: Window

__all__ = [
    "get_screens",
    "NoOpenGLException",
    "Window",
    "open_window",
    "View",
    "MOUSE_BUTTON_LEFT",
    "MOUSE_BUTTON_MIDDLE",
    "MOUSE_BUTTON_RIGHT",
]


def get_screens() -> list[Screen]:
    """
    Return a list of screens. So for a two-monitor setup, this should return
    a list of two screens. Can be used with :class:`arcade.Window` to select which
    window we full-screen on.

    Returns:
        List of screens, one for each monitor.
    """
    display = pyglet.display.get_display()
    return display.get_screens()


class NoOpenGLException(Exception):
    """Exception when we can't get an OpenGL 3.3+ context"""

    pass


class Window(pyglet.window.Window):
    """
    A window that will appear on your desktop.

    This class is a subclass of Pyglet's Window class with many
    Arcade-specific features added.

    .. note::

        Arcade currently cannot easily support multiple windows. If you need
        multiple windows, consider using multiple views or divide the window
        into sections.

    .. _pyglet_pg_window_size_position: https://pyglet.readthedocs.io/en/latest/programming_guide/windowing.html#size-and-position
    .. _pyglet_pg_window_style: https://pyglet.readthedocs.io/en/latest/programming_guide/windowing.html#window-style

    Args:
        width:
            Window width. Defaults to 1280.
        height:
            Window height. Defaults to 720.
        title:
            The title/caption of the window
        fullscreen:
            Should this be full screen?
        resizable:
            Can the user resize the window?
        update_rate:
            How frequently to run the on_update event.
        draw_rate:
            How frequently to run the on_draw event. (this is the FPS limit)
        fixed_rate:
            How frequently should the fixed_updates run,
            fixed updates will always run at this rate.
        fixed_frame_cap:
            The maximum number of fixed updates that can occur in one update loop.
            defaults to infinite. If large lag spikes cause your game to freeze,
            try setting this to a smaller number. This may cause your physics to
            lag behind temporarily.
        antialiasing:
            Use multisampling framebuffer (antialiasing)
        samples: Number of samples used in antialiasing (default 4).
            Usually this is 2, 4, 8 or 16.
        gl_version: What OpenGL version to request.
            This is ``(3, 3)`` by default and can be overridden when using more
            advanced OpenGL features.
        screen: Pass a pyglet :py:class:`~pyglet.display.Screen` to
            request the window be placed on it. See `pyglet's window size &
            position guide <pyglet_pg_window_size_position_>`_ to learn more.
        style: Request a non-default window style, such as borderless.
            Some styles only work in certain situations. See `pyglet's guide
            to window style <pyglet_pg_window_style_>`_ to learn more.
        visible:
            Should the window be visible immediately
        vsync:
            Wait for vertical screen refresh before swapping buffer
            This can make animations and movement look smoother.
        gc_mode: Decides how OpenGL objects should be garbage collected
            ("context_gc" (default) or "auto")
        center_window:
            If true, will center the window.
        enable_polling:
            Enabled input polling capability.
            This makes the :py:attr:`keyboard` and :py:attr:`mouse` attributes available for use.

    Raises:
        NoOpenGLException: If the system does not support OpenGL requested OpenGL version.
    """

    def __init__(
        self,
        width: int = 1280,
        height: int = 720,
        title: str | None = "Arcade Window",
        fullscreen: bool = False,
        resizable: bool = False,
        update_rate: float = 1 / 60,
        antialiasing: bool = True,
        gl_version: tuple[int, int] = (3, 3),
        screen: Screen | None = None,
        style: str | None = pyglet.window.Window.WINDOW_STYLE_DEFAULT,
        visible: bool = True,
        vsync: bool = False,
        gc_mode: str = "context_gc",
        center_window: bool = False,
        samples: int = 4,
        enable_polling: bool = True,
        gl_api: str = "gl",
        draw_rate: float = 1 / 60,
        fixed_rate: float = 1.0 / 60.0,
        fixed_frame_cap: int | None = None,
    ) -> None:
        # In certain environments we can't have antialiasing/MSAA enabled.
        # Detect replit environment
        if os.environ.get("REPL_ID"):
            antialiasing = False

        # Detect Raspberry Pi and switch to OpenGL ES 3.1
        if is_raspberry_pi():
            gl_version = 3, 1
            gl_api = "gles"

        self.headless: bool = arcade.headless
        """If True, the window is running in headless mode."""

        config = None
        # Attempt to make window with antialiasing
        if antialiasing:
            try:
                config = gl.Config(
                    major_version=gl_version[0],
                    minor_version=gl_version[1],
                    opengl_api=gl_api,  # type: ignore  # pending: upstream fix
                    double_buffer=True,
                    sample_buffers=1,
                    samples=samples,
                    depth_size=24,
                    stencil_size=8,
                    red_size=8,
                    green_size=8,
                    blue_size=8,
                    alpha_size=8,
                )
                display = pyglet.display.get_display()
                screen = screen or display.get_default_screen()
                if screen:
                    config = screen.get_best_config(config)
            except pyglet.window.NoSuchConfigException:
                LOG.warning("Skipping antialiasing due missing hardware/driver support")
                config = None
                antialiasing = False
        # If we still don't have a config
        if not config:
            config = gl.Config(
                major_version=gl_version[0],
                minor_version=gl_version[1],
                opengl_api=gl_api,  # type: ignore  # pending: upstream fix
                double_buffer=True,
                depth_size=24,
                stencil_size=8,
                red_size=8,
                green_size=8,
                blue_size=8,
                alpha_size=8,
            )
        try:
            super().__init__(
                width=width,
                height=height,
                caption=title,
                resizable=resizable,
                config=config,
                vsync=vsync,
                visible=visible,
                style=style,
            )
            # pending: weird import tricks resolved
            self.register_event_type("on_update")
            self.register_event_type("on_action")
            self.register_event_type("on_fixed_update")
        except pyglet.window.NoSuchConfigException:
            raise NoOpenGLException(
                "Unable to create an OpenGL 3.3+ context. "
                "Check to make sure your system supports OpenGL 3.3 or higher."
            )
        if antialiasing:
            try:
                gl.glEnable(gl.GL_MULTISAMPLE_ARB)
            except gl.GLException:
                LOG.warning("Warning: Anti-aliasing not supported on this computer.")

        _setup_clock()
        _setup_fixed_clock(fixed_rate)

        # We don't call the set_draw_rate function here because unlike the updates,
        # the draw scheduling is initially set in the call to pyglet.app.run()
        # that is done by the run() function. run() will pull this draw rate from
        # the Window and use it. Calls to set_draw_rate only need
        # to be done if changing it after the application has been started.

        # To ensure that draws are never de-synced from updates and wasted the draw rate
        # is forced to be slower than or equal to the update rate.
        # This works because pyglet ensures that a scheduled event takes as long or longer than the
        # call rate, but never less.
        assert update_rate <= draw_rate, (
            "An arcade window's draw rate cannot be faster than its update rate"
        )
        self._draw_rate = max(update_rate, draw_rate)
        self._accumulated_draw_time: float = 0.0

        # Fixed rate cannot be changed post initialization as this throws off physics sims.
        # If more time resolution is needed in fixed updates, devs can do 'sub-stepping'.
        self._fixed_rate = fixed_rate
        self._fixed_frame_cap = fixed_frame_cap
        self.set_update_rate(update_rate)

        self.set_vsync(vsync)

        if fullscreen is True:
            super().set_fullscreen(True, screen)

        set_window(self)

        self.push_handlers(on_resize=self._on_resize)

        self._ctx: ArcadeContext = ArcadeContext(self, gc_mode=gc_mode, gl_api=gl_api)
        self._background_color: Color = BLACK

        self._current_view: View | None = None

        # See if we should center the window
        if center_window:
            self.center_window()

        self.keyboard: pyglet.window.key.KeyStateHandler | None = None
        """
        A pyglet KeyStateHandler that can be used to poll the state of the keyboard.

            Example::

                    if self.window.keyboard[key.SPACE]:
                        print("The space key is currently being held down.")
        """
        self.mouse: pyglet.window.mouse.MouseStateHandler | None = None
        """
        A pyglet MouseStateHandler that can be used to poll the state of the mouse.

            Example::

                if self.window.mouse.LEFT:
                    print("The left mouse button is currently being held down.")
                print(
                    "The mouse is at position "
                    f"{self.window.mouse["x"]}, {self.window.mouse["y"]}"
                )
        """

        if enable_polling:
            self.keyboard = pyglet.window.key.KeyStateHandler()

            if arcade.headless:
                self.push_handlers(self.keyboard)

            else:
                self.mouse = pyglet.window.mouse.MouseStateHandler()
                self.push_handlers(self.keyboard)
                self.push_handlers(self.mouse)
        else:
            self.keyboard = None
            self.mouse = None

        # Framebuffer for drawing content into when start_render is called.
        # These are typically functions just at module level wrapped in
        # start_render and finish_render calls. The framebuffer is repeatedly
        # rendered to the window when the event loop starts.
        self._start_finish_render_data: StartFinishRenderData | None = None

    @property
    def current_view(self) -> View | None:
        """
        The currently active view.

        To set a different view, call :py:meth:`~arcade.Window.show_view`.
        """
        return self._current_view

    @property
    def ctx(self) -> ArcadeContext:
        """
        The OpenGL context for this window.

        This context instance provides access to a powerful set of
        features for lower level OpenGL programming. It is also used
        internally by Arcade to manage OpenGL resources.
        """
        return self._ctx

    def clear(  # type: ignore # not sure what to do here, BaseWindow.clear is static
        self,
        color: RGBOrA255 | None = None,
        color_normalized: RGBANormalized | None = None,
        viewport: tuple[int, int, int, int] | None = None,
    ) -> None:
        """
        Clears the window with the configured background color
        set through :py:attr:`~arcade.Window.background_color`.

        Args:
            color:
                Override the current background color with one of the following:

                1. A :py:class:`~arcade.types.Color` instance
                2. A 3 or 4-length RGB/RGBA :py:class:`tuple` of byte values (0 to 255)

            color_normalized:
                override the current background color using normalized values (0.0 to 1.0).
                For example, (1.0, 0.0, 0.0, 1.0) making the window contents red.

            viewport:
                The area of the window to clear. By default, the entire window is cleared.
                The viewport format is ``(x, y, width, height)``.
        """
        # Use the configured background color if none is provided
        if color is None and color_normalized is None:
            color = self.background_color
        self.ctx.screen.clear(color=color, color_normalized=color_normalized, viewport=viewport)

    @property
    def background_color(self) -> Color:
        """
        Get or set the background color for this window.
        This affects what color the window will contain when
        :py:meth:`~arcade.Window.clear` is called.

        Examples::

            # Use Arcade's built in Color values
            window.background_color = arcade.color.AMAZON

            # Set the background color with a custom Color instance
            MY_RED = arcade.types.Color(255, 0, 0)
            window.background_color = MY_RED

            # Set the background color directly from an RGBA tuple
            window.background_color = 255, 0, 0, 255

            # Set the background color directly from an RGB tuple
            # RGB tuples will assume 255 as the opacity / alpha value
            window.background_color = 255, 0, 0
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value: RGBOrA255) -> None:
        self._background_color = Color.from_iterable(value)

    @property
    def rect(self) -> Rect:
        """Return a Rect describing the size of the window."""
        return LBWH(0, 0, self.width, self.height)

    def run(self, view: View | None = None) -> None:
        """
        Run the event loop. Optionally start with a specified view.

        After the window has been set up, and the event hooks are in place, this
        is usually one of the last commands on the main program. This is a blocking
        function starting pyglet's event loop meaning it will start to dispatch
        events such as ``on_draw`` and ``on_update``.

        Args:
            view: The view to display when starting the run. Defaults to None.
        """
        if view is not None:
            self.show_view(view)
        arcade.run()

    def close(self) -> None:
        """Close the Window."""
        super().close()
        # Make sure we don't reference the window any more
        set_window(None)
        pyglet.clock.unschedule(self._dispatch_updates)

    def set_fullscreen(
        self,
        fullscreen: bool = True,
        screen=None,
        mode: ScreenMode | None = None,
        width: float | None = None,
        height: float | None = None,
    ) -> None:
        """
        Change the fullscreen status of the window.

        In most cases you simply want::

            # Enter fullscreen mode
            window.set_fullscreen(True)
            # Leave fullscreen mode
            window.set_fullscreen(False)

        When entering fullscreen mode the window will resize to the screen's
        resolution. When leaving fullscreen mode the window will resize back
        to the size it was before entering fullscreen mode.

        Args:
            fullscreen:
                Should we enter or leave fullscreen mode?
            screen:
                Which screen should we display on? See :func:`get_screens`
            mode:
                The screen will be switched to the given mode.  The mode must
                have been obtained by enumerating `Screen.get_modes`.  If
                None, an appropriate mode will be selected from the given
                `width` and `height`.
            width:
                Override the width of the window. Will be rounded to :py:class:`int`.
            height:
                Override the height of the window. Will be rounded to :py:class:`int`.
        """
        # fmt: off
        super().set_fullscreen(
            fullscreen, screen, mode,
            # TODO: resolve the upstream int / float screen coord issue
            None if width is None else int(width),
            None if height is None else int(height))
        # fmt: on

    def center_window(self) -> None:
        """Center the window on your desktop."""
        # Get the display screen using pyglet
        screen_width, screen_height = get_display_size()

        window_width, window_height = self.get_framebuffer_size()
        # Center the window
        self.set_location((screen_width - window_width) // 2, (screen_height - window_height) // 2)

    def on_update(self, delta_time: float) -> bool | None:
        """
        This method can be implemented and is reserved for game logic.
        Move sprites. Perform collision checks and other game logic.
        This method is called every frame before :meth:`on_draw`.

        The ``delta_time`` can be used to make sure the game runs at the same
        speed, no matter the frame rate.

        Args:
            delta_time: Time interval since the last time the function was
                called in seconds.
        """
        pass

    def on_fixed_update(self, delta_time: float):
        """
        Called for each fixed update. This is useful for physics engines
        and other systems that should update at a constant rate.

        Args:
            delta_time: Time interval since the last time the function was
                called in seconds.
        """
        pass

    def _dispatch_frame(self, delta_time: float) -> None:
        """
        To handle the de-syncing of on_draw and on_update that can occur when the events aren't
        linked. Dispatch frame keeps them in sync by always ensuring on_draw happens along-side
        an on_update. This requires that the draw frequencies is less than or equal to the update
        frequency.

        This only works because pyglet will only dispatch events after the call rate, or longer.
        This means if the update rate and draw rate are equal they will both always be called.
        The modulus on the accumulated draw time means that when the update rate is greater
        than the draw rate no time is lost.

        Args:
            delta_time: The amount of time since the last update.
        """
        self._dispatch_updates(delta_time)
        self._accumulated_draw_time += delta_time

        if self._draw_rate <= self._accumulated_draw_time:
            # Because we only ever dispatch one draw event per loop
            # we only need the modulus to keep time, if we didn't care
            # it could be set to zero instead.
            # ! This should maybe be fixed at 'self._draw_rate', discuss.
            self.draw(self._accumulated_draw_time)
            self._accumulated_draw_time %= self._draw_rate

    def _dispatch_updates(self, delta_time: float) -> None:
        """
        Internal function that is scheduled with Pyglet's clock, this function gets
        run by the clock, and dispatches the on_update events.

        It also accumulates time and runs fixed updates until the Fixed Clock catches
        up to the global clock

        Args:
            delta_time: Time interval since the last time the function was
                called in seconds.
        """
        GLOBAL_CLOCK.tick(delta_time)
        fixed_count = 0
        while GLOBAL_FIXED_CLOCK.accumulated >= self._fixed_rate and (
            self._fixed_frame_cap is None or fixed_count <= self._fixed_frame_cap
        ):
            GLOBAL_FIXED_CLOCK.tick(self._fixed_rate)
            self.dispatch_event("on_fixed_update", self._fixed_rate)
            fixed_count += 1
        self.dispatch_event("on_update", GLOBAL_CLOCK.delta_time)

    def flip(self) -> None:
        """
        Present the rendered content to the screen.

        This is not necessary to call when using the standard standard
        event loop. The event loop will automatically call this method
        after ``on_draw`` has been called.

        Window framebuffers normally have a back and front buffer meaning
        they are "double buffered". Content is always drawn into the back
        buffer while the front buffer contains the previous frame.
        Swapping the buffers makes the back buffer visible and hides the
        front buffer. This is done to prevent flickering and tearing.

        This method also garbage collects OpenGL resources if there are
        any dead resources to collect. If you override this method, make
        sure to call the super method to ensure that the garbage collection
        is done.
        """
        # Garbage collect OpenGL resources
        num_collected = self.ctx.gc()  # noqa: F841
        # LOG.debug("Garbage collected %s OpenGL resource(s)", num_collected)

        super().flip()  # type: ignore # Window typed at runtime

    def set_update_rate(self, rate: float) -> None:
        """
        Set how often the on_update function should be dispatched.
        For example::

            # Set the update rate to 60 times per second.
            self.set_update_rate(1 / 60)

        Args:
            rate: Update frequency in seconds
        """
        self._update_rate = rate
        pyglet.clock.unschedule(self._dispatch_frame)
        pyglet.clock.schedule_interval(self._dispatch_frame, rate)

    def set_draw_rate(self, rate: float) -> None:
        """
        Set how often the on_draw function should be run.
        The draw rate cannot currently be faster than the update rate.

        For example::

            # Set the draw rate to 60 frames per second.
            set.set_draw_rate(1 / 60)
        """
        assert self._update_rate <= rate, (
            "An arcade window's draw rate cannot be faster than its update rate"
        )
        self._draw_rate = max(self._update_rate, rate)

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> EVENT_HANDLE_STATE:
        """
        Called repeatedly while the mouse is moving in the window area.

        Override this function to respond to changes in mouse position.

        Args:
            x: x position of mouse within the window in pixels
            y: y position of mouse within the window in pixels
            dx: Change in x since the last time this method was called
            dy: Change in y since the last time this method was called
        """
        pass

    def on_mouse_press(self, x: int, y: int, button: int, modifiers: int) -> EVENT_HANDLE_STATE:
        """
        Called once whenever a mouse button gets pressed down.

        Override this function to handle mouse clicks. For an example of
        how to do this, see Arcade's built-in :ref:`aiming and shooting
        bullets <sprite_bullets_aimed>` demo.

        Args:
            x:
                x position of the mouse
            y:
                y position of the mouse
            button:
                What button was pressed. This will always be one of the following:

                - ``arcade.MOUSE_BUTTON_LEFT``
                - ``arcade.MOUSE_BUTTON_RIGHT``
                - ``arcade.MOUSE_BUTTON_MIDDLE``

            modifiers:
                Bitwise 'and' of all modifiers (shift, ctrl, num lock)
                active during this event. See :ref:`keyboard_modifiers`.
        """
        pass

    def on_mouse_drag(
        self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int
    ) -> EVENT_HANDLE_STATE:
        """
        Called repeatedly while the mouse moves with a button down.

        Override this function to handle dragging.

        Args:
            x:
                x position of mouse
            y:
                y position of mouse
            dx:
                Change in x since the last time this method was called
            dy:
                Change in y since the last time this method was called
            buttons:
                Which button is pressed
            modifiers:
                Bitwise 'and' of all modifiers (shift, ctrl, num lock)
                active during this event. See :ref:`keyboard_modifiers`.
        """
        return self.on_mouse_motion(x, y, dx, dy)

    def on_mouse_release(self, x: int, y: int, button: int, modifiers: int) -> EVENT_HANDLE_STATE:
        """
        Called once whenever a mouse button gets released.

        Override this function to respond to mouse button releases. This
        may be useful when you want to use the duration of a mouse click
        to affect gameplay.

        Args:
            x:
                x position of mouse
            y:
                y position of mouse
            button:
                What button was hit. One of:

                - ``arcade.MOUSE_BUTTON_LEFT``
                - ``arcade.MOUSE_BUTTON_RIGHT``
                - ``arcade.MOUSE_BUTTON_MIDDLE``
            modifiers:
                Bitwise 'and' of all modifiers (shift, ctrl, num lock)
                active during this event. See :ref:`keyboard_modifiers`.
        """
        return EVENT_UNHANDLED

    def on_mouse_scroll(
        self, x: int, y: int, scroll_x: float, scroll_y: float
    ) -> EVENT_HANDLE_STATE:
        """
        Called repeatedly while a mouse scroll wheel moves.

        Override this function to respond to scroll events. The scroll
        arguments may be positive or negative to indicate direction, but
        the units are unstandardized. How many scroll steps you receive
        may vary wildly between computers depending a number of factors,
        including system settings and the input devices used (i.e. mouse
        scrollwheel, touch pad, etc).

        .. warning:: Not all users can scroll easily!

                 Only some input devices support horizontal
                 scrolling. Standard vertical scrolling is common,
                 but some laptop touch pads are hard to use.

                 This means you should be careful about how you use
                 scrolling. Consider making it optional
                 to maximize the number of people who can play your
                 game!

        Args:
            x:
                x position of mouse
            y:
                y position of mouse
            scroll_x:
                Number of steps scrolled horizontally since the last call of this function
            scroll_y:
                Number of steps scrolled vertically since the last call of this function
        """
        return EVENT_UNHANDLED

    def set_mouse_visible(self, visible: bool = True) -> None:
        """
        Set whether to show the system's cursor while over the window

        By default, the system mouse cursor is visible whenever the
        mouse is over the window. To hide the cursor, pass ``False`` to
        this function. Pass ``True`` to make the cursor visible again.

        The window will continue receiving mouse events while the cursor
        is hidden, including movements and clicks. This means that
        functions like :meth:`~.Window.on_mouse_motion` and
        t':meth:`~.Window.on_mouse_press` will continue to work normally.

        You can use this behavior to visually replace the system mouse
        cursor with whatever you want. One example is :ref:`a game
        character that is always at the most recent mouse position in
        the window<sprite_collect_coins>`.

        .. note:: Advanced users can try using system cursor state icons

                 It may be possible to use system icons representing
                 cursor interaction states such as hourglasses or resize
                 arrows by using features :class:``~arcade.Window`` inherits
                 from the underlying pyglet window class. See the
                 `pyglet overview on cursors
                 <https://pyglet.readthedocs.io/en/master/programming_guide/mouse.html#changing-the-mouse-cursor>`_
                 for more information.

        Args:
            visible: Whether to hide the system mouse cursor
        """
        super().set_mouse_visible(visible)

    def on_action(self, action_name: str, state) -> None:
        """
        Called when an action is dispatched.
        This is related to the input manager / controller support.

        Args:
            action_name:
                The name of the action
            state:
                The state of the action
        """
        pass

    def on_key_press(self, symbol: int, modifiers: int) -> EVENT_HANDLE_STATE:
        """
        Called once when a key gets pushed down.

        Override this function to add key press functionality.

        .. tip:: If you want the length of key presses to affect
                 gameplay, you also need to override
                 :meth:`~.Window.on_key_release`.

        Args:
            symbol:
                Key that was just pushed down
            modifiers:
                Bitwise 'and' of all modifiers (shift, ctrl, num lock)
                active during this event. See :ref:`keyboard_modifiers`.
        """
        return EVENT_UNHANDLED

    def on_key_release(self, symbol: int, modifiers: int) -> EVENT_HANDLE_STATE:
        """
        Called once when a key gets released.

        Override this function to add key release functionality.

        Situations that require handling key releases include:

        * Rhythm games where a note must be held for a certain
          amount of time
        * 'Charging up' actions that change strength depending on
          how long a key was pressed
        * Showing which keys are currently pressed down

        Args:
            symbol (int): Key that was released
            modifiers (int): Bitwise 'and' of all modifiers (shift,
                      ctrl, num lock) active during this event.
                      See :ref:`keyboard_modifiers`.
        """
        return EVENT_UNHANDLED

    def on_draw(self) -> EVENT_HANDLE_STATE:
        """
        Override this function to add your custom drawing code.

        This method is usually called 60 times a second unless
        another update rate has been set. Should be called after
        :meth:`~arcade.Window.on_update`.

        This function should normally start with a call to
        :meth:`~arcade.Window.clear` to clear the screen.
        """
        if self._start_finish_render_data:
            self.clear()
            self._start_finish_render_data.draw()
            return True

        return EVENT_UNHANDLED

    def _on_resize(self, width: int, height: int) -> EVENT_HANDLE_STATE:
        """
        The internal method called when the window is resized.

        The purpose of this method is mainly setting the viewport
        to the new size of the window. Users should override
        :meth:`~arcade.Window.on_resize` instead. This method is
        called first.

        Args:
            width: New width of the window
            height: New height of the window
        """
        # Retain viewport
        self.viewport = (0, 0, width, height)

        return EVENT_UNHANDLED

    def on_resize(self, width: int, height: int) -> EVENT_HANDLE_STATE:
        """
        Override this method to add custom actions when the window is resized.

        An internal ``_on_resize`` is called first adjusting the viewport
        to the new size of the window so there is no need to call
        ```super().on_resize(width, height)```.

        Args:
            width: New width of the window
            height: New height of the window
        """
        pass

    def set_minimum_size(self, width: int, height: int) -> None:
        """
        Set the minimum size of the window.

        This will limit how small the window can be resized.

        Args:
            width: Minimum width
            height: Minimum height
        """
        super().set_minimum_size(width, height)

    def set_maximum_size(self, width: int, height: int) -> None:
        """
        Sets the maximum size of the window.

        This will limit how large the window can be resized.

        Args:
            width: Maximum width
            height: Maximum height
        """
        super().set_maximum_size(width, height)

    def set_size(self, width: int, height: int) -> None:
        """
        Resize the window.

        Args:
            width: New width of the window
            height: New height of the window
        """
        super().set_size(width, height)

    def get_size(self) -> tuple[int, int]:
        """Get the size of the window."""
        return super().get_size()

    def get_location(self) -> tuple[int, int]:
        """Get the current X/Y coordinates of the window."""
        return super().get_location()  # type: ignore # Window typed at runtime

    def set_visible(self, visible: bool = True):
        """
        Set if the window should be visible or not.

        Args:
            visible (bool): Should the window be visible?
        """
        super().set_visible(visible)

    def use(self) -> None:
        """Make the window the target for drawing.

        The window will always be the target for drawing unless
        offscreen framebuffers are used in the application.

        This simply binds the window's framebuffer.
        """
        self.ctx.screen.use()

    @property
    def default_camera(self) -> DefaultProjector:
        """
        The default camera for the window.

        This is an extremely simple camera simply responsible for
        maintaining the default projection and viewport.
        """
        return self._ctx._default_camera

    @property
    def current_camera(self) -> Projector:
        """
        Get or set the current camera.

        This represents the projector currently being used to define
        the projection and view matrices.
        """
        return self._ctx.current_camera

    @current_camera.setter
    def current_camera(self, next_camera):
        self._ctx.current_camera = next_camera

    @property
    def viewport(self) -> tuple[int, int, int, int]:
        """
        Get/set the viewport of the window.

        This will define what area of the window is rendered into.
        The values are ``x, y, width, height``. The value will normally
        be ``(0, 0, screen width, screen height)``.

        In most case you don't want to change this value manually
        and instead rely on the cameras.
        """
        return self._ctx.screen.viewport

    @viewport.setter
    def viewport(self, new_viewport: tuple[int, int, int, int]):
        if self._ctx.screen == self._ctx.active_framebuffer:
            self._ctx.viewport = new_viewport
        else:
            self._ctx.screen.viewport = new_viewport

    def test(self, frames: int = 10) -> None:
        """
        Used by unit test cases. Runs the event loop a few times and stops.

        Args:
            frames: How many frames to run the event loop for.
        """
        start_time = time.time()
        for _ in range(frames):
            self.switch_to()
            self.dispatch_events()
            self.dispatch_event("on_draw")
            self.flip()
            current_time = time.time()
            elapsed_time = current_time - start_time
            start_time = current_time
            if elapsed_time < 1.0 / 60.0:
                sleep_time = (1.0 / 60.0) - elapsed_time
                time.sleep(sleep_time)
            self._dispatch_updates(1 / 60)

    def show_view(self, new_view: View) -> None:
        """
        Set the currently active view.

        This will hide the current view
        and show the new view in the next frame.

        This is not a blocking call. It will simply point to the new view
        and return immediately.

        Calling this function is the same as setting the
        :py:attr:`arcade.Window.current_view` attribute.

        Args:
            new_view: The view to activate.
        """
        if not isinstance(new_view, View):
            raise TypeError(
                f"Window.show_view() takes an arcade.View, but it got a {type(new_view)}."
            )

        self._ctx.screen.use()
        self.viewport = (0, 0, self.width, self.height)

        # Store the Window that is showing the "new_view" View.
        if new_view.window is None:
            new_view.window = self
        # NOTE: This is not likely to happen and is creating issues for the test suite.
        # elif new_view.window != self:
        #     raise RuntimeError((
        #         "You are attempting to pass the same view "
        #         "object between multiple windows. A single "
        #         "view object can only be used in one window. "
        #         f"{self} != {new_view.window}"
        #     ))

        # remove previously shown view's handlers
        if self._current_view is not None:
            self._current_view.on_hide_view()
            self.remove_handlers(self._current_view)

        # push new view's handlers
        self._current_view = new_view

        # Note: Excluding on_show because this even can trigger multiple times.
        #       It should only be called once when the view is shown.
        view_handlers = {
            event_type: getattr(new_view, event_type, None)
            for event_type in self.event_types
            if event_type != "on_show" and hasattr(new_view, event_type)
        }
        if view_handlers:
            self.push_handlers(**view_handlers)
        self._current_view.on_show_view()

        # Note: After the View has been pushed onto pyglet's stack of event handlers
        # (via push_handlers()), pyglet
        # will still call the Window's event handlers.
        # (See pyglet's EventDispatcher.dispatch_event() implementation for details)

    def hide_view(self) -> None:
        """
        Hide the currently active view (if any).

        This is only necessary if you don't want an active view
        falling back to the window's event handlers. It's not
        necessary to call when changing the active view.
        """
        if self._current_view is None:
            return

        self._current_view.on_hide_view()
        self.remove_handlers(self._current_view)
        self._current_view = None

    def switch_to(self) -> None:
        """Switch the this window context.

        This is normally only used in multi-window applications.
        """
        super().switch_to()  # type: ignore # Window typed at runtime

    def set_caption(self, caption) -> None:
        """Set the caption/title of the window."""
        super().set_caption(caption)  # type: ignore # Window typed at runtime

    def set_location(self, x, y) -> None:
        """Set location of the window."""
        super().set_location(x, y)  # type: ignore # Window typed at runtime

    def activate(self) -> None:
        """Activate this window."""
        super().activate()  # type: ignore # Window typed at runtime

    def minimize(self) -> None:
        """Minimize the window."""
        super().minimize()  # type: ignore # Window typed at runtime

    def maximize(self) -> None:
        """Maximize  the window."""
        super().maximize()  # type: ignore # Window typed at runtime

    def set_vsync(self, vsync: bool) -> None:
        """Set if we sync our draws to the monitors vertical sync rate."""
        super().set_vsync(vsync)

    def set_mouse_platform_visible(self, platform_visible=None) -> None:
        """
        .. warning:: You are probably looking for
                     :meth:`~.Window.set_mouse_visible`!

        This is a lower level function inherited from the pyglet window.

        For more information on what this means, see the documentation
        for :py:meth:`pyglet.window.Window.set_mouse_platform_visible`.
        """
        super().set_mouse_platform_visible(platform_visible)

    def set_exclusive_mouse(self, exclusive=True) -> None:
        """Capture the mouse."""
        super().set_exclusive_mouse(exclusive)

    def set_exclusive_keyboard(self, exclusive=True) -> None:
        """Capture all keyboard input."""
        super().set_exclusive_keyboard(exclusive)

    def get_system_mouse_cursor(self, name) -> MouseCursor:
        """Get the system mouse cursor"""
        return super().get_system_mouse_cursor(name)

    def dispatch_events(self) -> None:
        """Dispatch events"""
        super().dispatch_events()  # type: ignore # Window typed at runtime

    def on_mouse_enter(self, x: int, y: int) -> EVENT_HANDLE_STATE:
        """
        Called once whenever the mouse enters the window area on screen.

        This event will not be triggered if the mouse is currently being
        dragged.

        Args:
            x: The x position the mouse entered the window
            y: The y position the mouse entered the window
        """
        pass

    def on_mouse_leave(self, x: int, y: int) -> EVENT_HANDLE_STATE:
        """
        Called once whenever the mouse leaves the window area on screen.

        This event will not be triggered if the mouse is currently being
        dragged. Note that the coordinates of the mouse pointer will be
        outside of the window rectangle.

        Args:
            x: The x position the mouse entered the window
            y: The y position the mouse entered the window
        """
        pass

    @property
    def center(self) -> tuple[float, float]:
        """
        Returns center coordinates of the window

        Equivalent to ``(self.width / 2, self.height / 2)``.
        """
        return (self.width / 2, self.height / 2)

    @property
    def center_x(self) -> float:
        """
        Returns the center x-coordinate of the window.

        Equivalent to ``self.width / 2``.
        """
        return self.width / 2

    @property
    def center_y(self) -> float:
        """
        Returns the center y-coordinate of the window.

        Equivalent to ``self.height / 2``.
        """
        return self.height / 2

    # --- CLOCK ALIASES ---
    @property
    def time(self) -> float:
        """
        Shortcut to the global clock's time.

        This is the time in seconds since the application started.
        """
        return GLOBAL_CLOCK.time

    @property
    def fixed_time(self) -> float:
        """
        Shortcut to the fixed clock's time.

        This is the time in seconds since the application started
        but updated at a fixed rate.
        """
        return GLOBAL_FIXED_CLOCK.time

    @property
    def delta_time(self) -> float:
        """Shortcut for the global clock's delta_time."""
        return GLOBAL_CLOCK.delta_time

    @property
    def fixed_delta_time(self) -> float:
        """The configured fixed update rate"""
        return self._fixed_rate

    # required because pyglet marks the method as abstract methods,
    # but resolves class during runtime
    def _create(self) -> None:
        """Internal method to create the window."""
        super()._create()  # type: ignore

    def _recreate(self, changes: Sequence[str]) -> None:
        super()._recreate(changes)  # type: ignore


def open_window(
    width: int,
    height: int,
    window_title: str | None = None,
    resizable: bool = False,
    antialiasing: bool = True,
    **kwargs,
) -> Window:
    """
    Shortcut for opening/creating a window with less options.

    For a full set of window options, create a :py:class:`~arcade.Window`
    instance directly.

    Args:
        width:
            Width of the window.
        height:
            Height of the window.
        window_title:
            Title/caption of the window.
        resizable:
            Whether the user can resize the window.
        antialiasing:
            Whether to use antialiasing
        **kwargs:
            Additional keyword arguments to pass to the window constructor.
    """
    global _window
    _window = Window(
        width, height, window_title, resizable=resizable, antialiasing=antialiasing, **kwargs
    )
    _window.invalid = False
    return _window


class View:
    """
    A view is a way to separate drawing and logic from the window itself.
    Subclassing the window is very inflexible since you can't easily switch
    your update and draw logic.

    A view is a way to encapsulate that logic, so you can easily switch between
    different parts of your game. Maybe you have a title screen, a game screen,
    and a game over screen. Each of these could be a different view.

    Args:
        window:
            The window this view is associated with. If None, the current
            window is used. (Normally you don't need to provide this).
    """

    def __init__(
        self, window: Window | None = None, background_color: RGBOrA255 | None = None
    ) -> None:
        self.window = arcade.get_window() if window is None else window
        self._background_color: Color | None = background_color and Color.from_iterable(
            background_color
        )

    def clear(
        self,
        color: RGBOrA255 | None = None,
        color_normalized: RGBANormalized | None = None,
        viewport: tuple[int, int, int, int] | None = None,
    ) -> None:
        """
        Clears the window with the configured background color
        set through :py:attr:`arcade.View.background_color`.

        Args:
            color:
                override the current background color with one of the following:

                1. A :py:class:`~arcade.types.Color` instance
                2. A 3 or 4-length RGB/RGBA :py:class:`tuple` of byte values (0 to 255)
            color_normalized:
                Override the current background color using normalized values (0.0 to 1.0).
                For example, (1.0, 0.0, 0.0, 1.0) making the window contents red.
            viewport:
                The viewport range to clear
        """
        if color is None and color_normalized is None:
            color = self.background_color
        self.window.clear(color=color, color_normalized=color_normalized, viewport=viewport)

    def on_update(self, delta_time: float) -> bool | None:
        """
        This method can be implemented and is reserved for game logic.
        Move sprites. Perform collision checks and other game logic.
        This method is called every frame before :meth:`on_draw`.

        The ``delta_time`` can be used to make sure the game runs at the same
        speed, no matter the frame rate.

        Args:
            delta_time:
                Time interval since the last time the function was called in seconds.
        """
        pass

    def on_fixed_update(self, delta_time: float):
        """
        Called for each fixed update. This is useful for physics engines
        and other systems that should update at a constant rate.

        Args:
            delta_time:
                Time interval since the last time the function was called in seconds.
        """
        pass

    def on_draw(self) -> bool | None:
        """
        Override this function to add your custom drawing code.

        This method is usually called 60 times a second unless
        another update rate has been set. Should be called after
        :meth:`~arcade.Window.on_update`.

        This function should normally start with a call to
        :meth:`~arcade.Window.clear` to clear the screen.
        """
        pass

    def on_show_view(self) -> None:
        """Called once when the view is shown.

        .. seealso:: :py:meth:`~arcade.View.on_hide_view`
        """
        pass

    def on_hide_view(self) -> None:
        """Called once when this view is hidden."""
        pass

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> bool | None:
        """
        Called repeatedly while the mouse is moving in the window area.

        Override this function to respond to changes in mouse position.

        Args:
            x: x position of mouse within the window in pixels
            y: y position of mouse within the window in pixels
            dx: Change in x since the last time this method was called
            dy: Change in y since the last time this method was called
        """
        pass

    def on_mouse_press(self, x: int, y: int, button: int, modifiers: int) -> bool | None:
        """
        Called once whenever a mouse button gets pressed down.

        Override this function to handle mouse clicks. For an example of
        how to do this, see Arcade's built-in :ref:`aiming and shooting
        bullets <sprite_bullets_aimed>` demo.

        Args:
            x:
                x position of the mouse
            y:
                y position of the mouse
            button:
                What button was pressed. This will always be one of the following:

                - ``arcade.MOUSE_BUTTON_LEFT``
                - ``arcade.MOUSE_BUTTON_RIGHT``
                - ``arcade.MOUSE_BUTTON_MIDDLE``

            modifiers:
                Bitwise 'and' of all modifiers (shift, ctrl, num lock)
                active during this event. See :ref:`keyboard_modifiers`.
        """
        pass

    def on_mouse_drag(
        self, x: int, y: int, dx: int, dy: int, _buttons: int, _modifiers: int
    ) -> bool | None:
        """
        Called repeatedly while the mouse moves with a button down.

        Override this function to handle dragging.

        Args:
            x:
                x position of mouse
            y:
                y position of mouse
            dx:
                Change in x since the last time this method was called
            dy:
                Change in y since the last time this method was called
            _buttons:
                Which button is pressed
            _modifiers:
                Bitwise 'and' of all modifiers (shift, ctrl, num lock)
                active during this event. See :ref:`keyboard_modifiers`.
        """
        self.on_mouse_motion(x, y, dx, dy)
        return False

    def on_mouse_release(self, x: int, y: int, button: int, modifiers: int) -> bool | None:
        """
        Called once whenever a mouse button gets released.

        Override this function to respond to mouse button releases. This
        may be useful when you want to use the duration of a mouse click
        to affect gameplay.

        Args:
            x:
                x position of mouse
            y:
                y position of mouse
            button:
                What button was hit. One of:

                - ``arcade.MOUSE_BUTTON_LEFT``
                - ``arcade.MOUSE_BUTTON_RIGHT``
                - ``arcade.MOUSE_BUTTON_MIDDLE``

            modifiers:
                Bitwise 'and' of all modifiers (shift, ctrl, num lock)
                active during this event. See :ref:`keyboard_modifiers`.
        """
        pass

    def on_mouse_scroll(self, x: int, y: int, scroll_x: int, scroll_y: int) -> bool | None:
        """
        Called repeatedly while a mouse scroll wheel moves.

        Override this function to respond to scroll events. The scroll
        arguments may be positive or negative to indicate direction, but
        the units are unstandardized. How many scroll steps you receive
        may vary wildly between computers depending a number of factors,
        including system settings and the input devices used (i.e. mouse
        scrollwheel, touch pad, etc).

        .. warning:: Not all users can scroll easily!

            Only some input devices support horizontal
            scrolling. Standard vertical scrolling is common,
            but some laptop touch pads are hard to use.

            This means you should be careful about how you use
            scrolling. Consider making it optional
            to maximize the number of people who can play your
            game!

        Args:
            x:
                x position of mouse
            y:
                y position of mouse
            scroll_x:
                number of steps scrolled horizontally
                since the last call of this function
            scroll_y:
                number of steps scrolled vertically since
                the last call of this function
        """
        pass

    def on_key_press(self, symbol: int, modifiers: int) -> bool | None:
        """
        Called once when a key gets pushed down.

        Override this function to add key press functionality.

        .. tip:: If you want the length of key presses to affect
                 gameplay, you also need to override
                 :meth:`~.Window.on_key_release`.

        Args:
            symbol:
                Key that was just pushed down
            modifiers:
                Bitwise 'and' of all modifiers (shift, ctrl, num lock) active
                during this event. See :ref:`keyboard_modifiers`.
        """
        return False

    def on_key_release(self, symbol: int, modifiers: int) -> bool | None:
        """
        Called once when a key gets released.

        Override this function to add key release functionality.

        Situations that require handling key releases include:

        * Rhythm games where a note must be held for a certain
          amount of time
        * 'Charging up' actions that change strength depending on
          how long a key was pressed
        * Showing which keys are currently pressed down

        Args:
            symbol:
                Key that was released
            modifiers:
                Bitwise 'and' of all modifiers (shift, ctrl, num lock) active
                during this event. See :ref:`keyboard_modifiers`.
        """
        return False

    def on_resize(self, width: int, height: int) -> bool | None:
        """
        Override this method to add custom actions when the window is resized.

        An internal ``_on_resize`` is called first adjusting the viewport
        to the new size of the window so there is no need to call
        ```super().on_resize(width, height)```.

        Args:
            width: New width of the window
            height: New height of the window
        """
        pass

    def on_mouse_enter(self, x: int, y: int) -> bool | None:
        """
        Called once whenever the mouse enters the window area on screen.

        This event will not be triggered if the mouse is currently being
        dragged.

        Args:
            x: The x position the mouse entered the window
            y: The y position the mouse entered the window
        """
        pass

    def on_mouse_leave(self, x: int, y: int) -> bool | None:
        """
        Called once whenever the mouse leaves the window area on screen.

        This event will not be triggered if the mouse is currently being
        dragged. Note that the coordinates of the mouse pointer will be
        outside of the window rectangle.

        Args:
            x: The x position the mouse entered the window
            y: The y position the mouse entered the window
        """
        pass

    @property
    def size(self) -> tuple[float, float]:
        """
        An alias for `arcade.Window.size`
        """
        return self.window.size

    @property
    def width(self) -> float:
        """
        An alias for `arcade.Window.width`
        """
        return self.window.width

    @property
    def height(self) -> float:
        """
        An alias for `arcade.Window.height`
        """
        return self.window.height

    @property
    def center(self) -> tuple[float, float]:
        """
        An alias for `arcade.Window.center`
        """
        return self.window.center

    @property
    def center_x(self) -> float:
        """
        An alias for `arcade.Window.center_x`
        """
        return self.window.center_x

    @property
    def center_y(self) -> float:
        """
        An alias for `arcade.Window.center_y`
        """
        return self.window.center_y

    @property
    def background_color(self) -> Color | None:
        """
        Get or set the background color for this view.
        This affects what color the window will contain when
        :py:meth:`~arcade.View.clear` is called.

        Examples::

            # Use Arcade's built in Color values
            view.background_color = arcade.color.AMAZON

            # Set the background color with a custom Color instance
            MY_RED = arcade.types.Color(255, 0, 0)
            view.background_color = MY_RED

            # Set the background color directly from an RGBA tuple
            view.background_color = 255, 0, 0, 255

            # Set the background color directly from an RGB tuple
            # RGB tuples will assume 255 as the opacity / alpha value
            view.background_color = 255, 0, 0
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value: RGBOrA255) -> None:
        self._background_color = Color.from_iterable(value)



"""
Section Example 3:

This shows how sections work with a very small example

What's key here is to understand how sections can isolate code that otherwise
 goes packed together in the view.
Also, note that events are received on each section only based on the
 section configuration. This way you don't have to check every time if the mouse
 position is on top of some area.

Note:
 - Event dispatching (two sections will receive on_key_press and on_key_release)
 - Prevent dispatching to allow some events to stop propagating
 - Event draw, update and event delivering order based on section_manager
   sections list order
 - Section "enable" property to show or hide sections
 - Modal Sections: sections that draw last but capture all events and also stop
   other sections from updating.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sections_demo_3
"""

from math import sqrt

import arcade
from arcade import Section, SectionManager
from arcade.types import Color

INFO_BAR_HEIGHT = 40
PANEL_WIDTH = 200
SPRITE_SPEED = 1

COLOR_LIGHT = Color.from_hex_string("#D9BBA0")
COLOR_DARK = Color.from_hex_string("#0D0D0D")
COLOR_1 = Color.from_hex_string("#2A1459")
COLOR_2 = Color.from_hex_string("#4B89BF")
COLOR_3 = Color.from_hex_string("#03A688")


class Ball(arcade.SpriteCircle):
    """The moving ball"""

    def __init__(self, radius, color):
        super().__init__(radius, color)

        self.bounce_count: int = 0  # to count the number of bounces

    @property
    def speed(self):
        # return euclidian distance * current fps (60 default)
        return int(sqrt(pow(self.change_x, 2) + pow(self.change_y, 2)) * 60)


class ModalSection(Section):
    """A modal section that represents a popup that waits for user input"""

    def __init__(self, left: int, bottom: int, width: int, height: int):
        super().__init__(left, bottom, width, height, modal=True, enabled=False)

        # modal button
        self.button = arcade.SpriteSolidColor(100, 50, color=arcade.color.RED)
        pos = self.left + self.width / 2, self.bottom + self.height / 2
        self.button.position = pos

    def on_draw(self):
        # draw modal frame and button
        arcade.draw_lrbt_rectangle_filled(
            self.left, self.right, self.bottom, self.top, arcade.color.GRAY
        )
        arcade.draw_lrbt_rectangle_outline(
            self.left, self.right, self.bottom, self.top, arcade.color.WHITE
        )
        self.draw_button()

    def draw_button(self):
        # draws the button and button text
        arcade.draw_sprite(self.button)
        arcade.draw_text(
            "Close Modal",
            self.button.left + 5,
            self.button.bottom + self.button.height / 2,
            arcade.color.WHITE,
        )

    def on_resize(self, width: int, height: int):
        """set position on screen resize"""
        self.left = width // 3
        self.bottom = (height // 2) - self.height // 2
        pos = self.left + self.width / 2, self.bottom + self.height / 2
        self.button.position = pos

    def on_mouse_press(self, x: float, y: float, button: int, modifiers: int):
        """Check if the button is pressed"""
        if self.button.collides_with_point((x, y)):
            self.enabled = False


class InfoBar(Section):
    """This is the top bar of the screen where info is showed"""

    @property
    def ball(self):
        return self.view.map.ball

    def on_draw(self):
        # draw game info
        arcade.draw_lrbt_rectangle_filled(self.left, self.right, self.bottom, self.top, COLOR_DARK)
        arcade.draw_lrbt_rectangle_outline(
            self.left, self.right, self.bottom, self.top, COLOR_LIGHT
        )
        arcade.draw_text(
            f"Ball bounce count: {self.ball.bounce_count}",
            self.left + 20,
            self.top - self.height / 1.6,
            COLOR_LIGHT,
        )

        ball_change_axis = self.ball.change_x, self.ball.change_y
        arcade.draw_text(
            f"Ball change in axis: {ball_change_axis}",
            self.left + 220,
            self.top - self.height / 1.6,
            COLOR_LIGHT,
        )
        arcade.draw_text(
            f"Ball speed: {self.ball.speed} pixels/second",
            self.left + 480,
            self.top - self.height / 1.6,
            COLOR_LIGHT,
        )

    def on_resize(self, width: int, height: int):
        # stick to the top
        self.width = width
        self.bottom = height - self.view.info_bar.height


class Panel(Section):
    """This is the Panel to the right where buttons and info is showed"""

    def __init__(self, left: int, bottom: int, width: int, height: int, **kwargs):
        super().__init__(left, bottom, width, height, **kwargs)

        # create buttons
        self.button_stop = self.new_button(arcade.color.ARSENIC)
        self.button_toggle_info_bar = self.new_button(COLOR_1)

        self.button_show_modal = self.new_button(COLOR_2)
        # to show the key that's actually pressed
        self.pressed_key: int | None = None

    @staticmethod
    def new_button(color):
        # helper to create new buttons
        return arcade.SpriteSolidColor(100, 50, color=color)

    def draw_button_stop(self):
        arcade.draw_text(
            "Press button to stop the ball", self.left + 10, self.top - 40, COLOR_LIGHT, 10
        )
        arcade.draw_sprite(self.button_stop)

    def draw_button_toggle_info_bar(self):
        arcade.draw_text(
            "Press to toggle info_bar", self.left + 10, self.top - 140, COLOR_LIGHT, 10
        )
        arcade.draw_sprite(self.button_toggle_info_bar)

    def draw_button_show_modal(self):
        arcade.draw_sprite(self.button_show_modal)
        arcade.draw_text(
            "Show Modal", self.left - 37 + self.width / 2, self.bottom + 95, COLOR_DARK, 10
        )

    def on_draw(self):
        arcade.draw_lrbt_rectangle_filled(self.left, self.right, self.bottom, self.top, COLOR_DARK)
        arcade.draw_lrbt_rectangle_outline(
            self.left, self.right, self.bottom, self.top, COLOR_LIGHT
        )
        self.draw_button_stop()
        self.draw_button_toggle_info_bar()

        if self.pressed_key:
            arcade.draw_text(
                f"Pressed key code: {self.pressed_key}",
                self.left + 10,
                self.top - 240,
                COLOR_LIGHT,
                9,
            )

        self.draw_button_show_modal()

    def on_mouse_press(self, x: float, y: float, button: int, modifiers: int):
        if self.button_stop.collides_with_point((x, y)):
            self.view.map.ball.stop()
        elif self.button_toggle_info_bar.collides_with_point((x, y)):
            self.view.info_bar.enabled = not self.view.info_bar.enabled
        elif self.button_show_modal.collides_with_point((x, y)):
            self.view.modal_section.enabled = True

    def on_resize(self, width: int, height: int):
        # stick to the right
        self.left = width - self.width
        self.height = height - self.view.info_bar.height
        self.button_stop.position = self.left + self.width / 2, self.top - 80

        pos = self.left + self.width / 2, self.top - 180
        self.button_toggle_info_bar.position = pos

        pos = self.left + self.width / 2, self.bottom + 100
        self.button_show_modal.position = pos

    def on_key_press(self, symbol: int, modifiers: int):
        self.pressed_key = symbol

    def on_key_release(self, _symbol: int, _modifiers: int):
        self.pressed_key = None


class Map(Section):
    """This represents the place where the game takes place"""

    def __init__(self, left: int, bottom: int, width: int, height: int, **kwargs):
        super().__init__(left, bottom, width, height, **kwargs)

        self.ball = Ball(20, COLOR_3)
        self.ball.position = 60, 60
        self.sprite_list: arcade.SpriteList = arcade.SpriteList()
        self.sprite_list.append(self.ball)

        self.pressed_key: int | None = None

    def on_update(self, delta_time: float):
        if self.pressed_key:
            if self.pressed_key == arcade.key.UP:
                self.ball.change_y += SPRITE_SPEED
            elif self.pressed_key == arcade.key.RIGHT:
                self.ball.change_x += SPRITE_SPEED
            elif self.pressed_key == arcade.key.DOWN:
                self.ball.change_y -= SPRITE_SPEED
            elif self.pressed_key == arcade.key.LEFT:
                self.ball.change_x -= SPRITE_SPEED

        self.sprite_list.update()

        if self.ball.top >= self.top or self.ball.bottom <= self.bottom:
            self.ball.change_y *= -1
            self.ball.bounce_count += 1
        if self.ball.left <= self.left or self.ball.right >= self.right:
            self.ball.change_x *= -1
            self.ball.bounce_count += 1

    def on_draw(self):
        arcade.draw_lrbt_rectangle_filled(self.left, self.right, self.bottom, self.top, COLOR_DARK)
        arcade.draw_lrbt_rectangle_outline(
            self.left, self.right, self.bottom, self.top, COLOR_LIGHT
        )
        self.sprite_list.draw()

    def on_key_press(self, symbol: int, modifiers: int):
        self.pressed_key = symbol

    def on_key_release(self, _symbol: int, _modifiers: int):
        self.pressed_key = None

    def on_resize(self, width: int, height: int):
        self.width = width - self.view.panel.width
        self.height = height - self.view.info_bar.height


class GameView(arcade.View):
    """The game itself"""

    def __init__(self):
        super().__init__()

        # create and store the modal, so we can set
        # self.modal_section.enabled = True to show it
        self.modal_section = ModalSection(
            (self.window.width / 2) - 150,
            (self.window.height / 2) - 100,
            300,
            200,
        )

        # we set accept_keyboard_events to False (default to True)
        self.info_bar = InfoBar(
            0,
            self.window.height - INFO_BAR_HEIGHT,
            self.window.width,
            INFO_BAR_HEIGHT,
            accept_keyboard_keys=False,
        )

        # as prevent_dispatch is on by default, we let pass the events to the
        # following Section: the map
        self.panel = Panel(
            self.window.width - PANEL_WIDTH,
            0,
            PANEL_WIDTH,
            self.window.height - INFO_BAR_HEIGHT,
            prevent_dispatch={False},
        )
        self.map = Map(0, 0, self.window.width - PANEL_WIDTH, self.window.height - INFO_BAR_HEIGHT)

        # add the sections
        self.section_manager = SectionManager(self)
        self.section_manager.add_section(self.modal_section)
        self.section_manager.add_section(self.info_bar)
        self.section_manager.add_section(self.panel)
        self.section_manager.add_section(self.map)

    def on_show_view(self) -> None:
        self.section_manager.enable()

    def on_hide_view(self) -> None:
        self.section_manager.disable()

    def on_draw(self):
        self.clear()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(resizable=True)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Convenience functions that provide a much simpler interface to Emitters and Particles.

These trade away some flexibility in favor of simplicity to allow beginners
to start using particle systems.
"""

import random
from collections.abc import Sequence

from arcade.math import rand_in_circle, rand_on_circle
from arcade.types import PathOrTexture, Point

from .emitter import EmitBurst, Emitter, EmitterIntervalWithTime
from .particle import FadeParticle, LifetimeParticle


def make_burst_emitter(
    center_xy: Point,
    filenames_and_textures: Sequence[PathOrTexture],
    particle_count: int,
    particle_speed: float,
    particle_lifetime_min: float,
    particle_lifetime_max: float,
    particle_scale: float = 1.0,
    fade_particles: bool = True,
) -> Emitter:
    """Returns an emitter that emits all of its particles at once"""
    particle_factory: type[LifetimeParticle] = LifetimeParticle
    if fade_particles:
        particle_factory = FadeParticle
    return Emitter(
        center_xy=center_xy,
        emit_controller=EmitBurst(particle_count),
        particle_factory=lambda emitter: particle_factory(
            filename_or_texture=random.choice(filenames_and_textures),
            change_xy=rand_in_circle((0.0, 0.0), particle_speed),
            lifetime=random.uniform(particle_lifetime_min, particle_lifetime_max),
            scale=particle_scale,
        ),
    )


def make_interval_emitter(
    center_xy: Point,
    filenames_and_textures: Sequence[PathOrTexture],
    emit_interval: float,
    emit_duration: float,
    particle_speed: float,
    particle_lifetime_min: float,
    particle_lifetime_max: float,
    particle_scale: float = 1.0,
    fade_particles: bool = True,
) -> Emitter:
    """Returns an emitter that emits its particles at a constant rate for a given amount of time"""
    particle_factory: type[LifetimeParticle] = LifetimeParticle
    if fade_particles:
        particle_factory = FadeParticle
    return Emitter(
        center_xy=center_xy,
        emit_controller=EmitterIntervalWithTime(emit_interval, emit_duration),
        particle_factory=lambda emitter: particle_factory(
            filename_or_texture=random.choice(filenames_and_textures),
            change_xy=rand_on_circle((0.0, 0.0), particle_speed),
            lifetime=random.uniform(particle_lifetime_min, particle_lifetime_max),
            scale=particle_scale,
        ),
    )



"""
Simple instancing example.

We draw a triangle N times using instancing.
A position offset and color is passed in per instance.
To make things a bit more interesting we also rotate the triangle.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.instancing
"""

import random
from array import array

import arcade
from arcade.gl import BufferDescription

WINDOW_WIDTH = 800
WINDOW_HEIGHT = 800
WINDOW_TITLE = "Instancing"


class GameView(arcade.Window):

    def __init__(self, width, height, title):
        """
        Set up the application.
        """
        super().__init__(width, height, title)
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            uniform float time;

            in vec2 in_vert;
            // Per instance data
            in vec2 in_offset;
            in vec4 in_color;

            out vec4 v_color;

            void main() {
                // Create rotation based on what instance we are drawing
                float angle = time * 0.5 + float(gl_InstanceID);
                mat2 rot = mat2(
                    cos(angle), sin(angle),
                    -sin(angle), cos(angle)
                );

                // scale down triangle, offset and rotate
                vec2 pos = (rot * (in_vert * 0.07)) + in_offset;

                gl_Position = vec4(pos, 0.0, 1.0);
                v_color = in_color;
            }
            """,
            fragment_shader="""
            #version 330

            in vec4 v_color;
            out vec4 out_color;

            void main() {
                out_color = v_color;
            }
            """,
        )

        self.instances = 1_000
        # Create triangle
        # fmt: off
        vertices = array("f", [
            # x, y
            0.0, 0.8,
            -0.8, -0.8,
            0.8, -0.8,
        ])
        # fmt: on

        # Generate per instance data. We'll create a generator function for this (less messy)
        def gen_instance_data(instances):
            random.seed(123456)
            for _ in range(instances):
                # random x and y
                yield random.uniform(-1.0, 1.0)
                yield random.uniform(-1.0, 1.0)
                # random rgba color
                yield random.uniform(0.1, 1.0)
                yield random.uniform(0.1, 1.0)
                yield random.uniform(0.1, 1.0)
                yield random.uniform(0.1, 1.0)

        per_instance = array("f", gen_instance_data(self.instances))

        self.geometry = self.ctx.geometry(
            [
                # Base geometry
                BufferDescription(
                    self.ctx.buffer(data=vertices),
                    "2f",
                    ["in_vert"],
                ),
                # Per instance buffer
                BufferDescription(
                    self.ctx.buffer(data=per_instance),
                    "2f 4f",
                    ["in_offset", "in_color"],
                    instanced=True,
                ),
            ],
            mode=self.ctx.TRIANGLES,
        )

    def on_draw(self):
        self.clear()
        self.ctx.enable(self.ctx.BLEND)

        self.program["time"] = self.time
        self.geometry.render(self.program, instances=self.instances)


if __name__ == "__main__":
    GameView(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    arcade.run()



"""Sound Library."""

import logging
import math
import os
from pathlib import Path

import pyglet
from pyglet.media import Source

from arcade.resources import resolve

if os.environ.get("ARCADE_SOUND_BACKENDS"):
    pyglet.options.audio = tuple(v.strip() for v in os.environ["ARCADE_SOUND_BACKENDS"].split(","))
else:
    pyglet.options.audio = ("openal", "xaudio2", "directsound", "pulse", "silent")

import pyglet.media as media

__all__ = ["Sound", "load_sound", "play_sound", "stop_sound"]

logger = logging.getLogger("arcade")


class Sound:
    """Holds :ref:`playable <sound-basics-playing>` loaded audio data.

    .. important:: :ref:`Streaming <sound-loading-modes>` disables features!

    When ``streaming=True``, :py:meth:`.play` and :py:func:`play_sound`:

    * raise a :py:class:`RuntimeError` if there is already another
      active playback
    * do not support looping

    To learn about the restrictions on :ref:`streaming <sound-loading-modes>`,
    please see:

    * :ref:`sound-loading-modes-streaming`
    * The :py:class:`pyglet.media.codecs.base.StreamingSource` class used
      internally

    To learn about cross-platform loading and file format concerns,
    please see:

    * Arcade's sound documentation:

      * :ref:`sound-loading-modes`
      * :ref:`sound-compat-easy`
      * :ref:`sound-compat-loading`

    * The pyglet guide to :external+pyglet:ref:`guide-media`

    Args:
         file_name:
            The path of a file to load, optionally prefixed with a
            :ref:`resource handle <resource_handles>`.
         streaming:
            If ``True``, attempt to load data from ``file_path`` via
            via :ref:`streaming <sound-loading-modes>`.
    """

    def __init__(self, file_name: str | Path, streaming: bool = False):
        self.file_name: str = ""
        file_name = resolve(file_name)

        if not Path(file_name).is_file():
            raise FileNotFoundError(f"The sound file '{file_name}' is not a file or can't be read.")
        self.file_name = str(file_name)

        self.source: Source = media.load(self.file_name, streaming=streaming)
        """
        The :py:class:`pyglet.media.Source` object that holds the audio data.
        """

        if self.source.duration is None:
            raise ValueError(
                "Audio duration must be known when loaded, but this audio source returned `None`"
            )

        self.min_distance = (
            100000000  # setting the players to this allows for 2D panning with 3D audio
        )

    def play(
        self,
        volume: float = 1.0,
        pan: float = 0.0,
        loop: bool = False,
        speed: float = 1.0,
    ) -> media.Player:
        """Try to play this :py:class:`Sound` and return a |pyglet Player|.

        .. important:: A :py:class:`Sound` with ``streaming=True`` loses features!

                       Neither ``loop`` nor simultaneous playbacks will work. See
                       :py;class:`Sound` and :ref:`sound-loading-modes`.

        Args:
            volume: Volume (``0.0`` is silent, ``1.0`` is loudest).
            pan: Left / right channel balance (``-1`` is left,  ``0.0`` is
                center, and ``1.0`` is right).
            loop: ``True`` attempts to restart playback after finishing.
            speed: Change the speed (and pitch) of the sound. Default speed is
                ``1.0``.
        Returns:
            A |pyglet Player| for this playback.
        """
        if isinstance(self.source, media.StreamingSource) and self.source.is_player_source:
            raise RuntimeError(
                "Tried to play a streaming source more than once."
                " Streaming sources should only be played in one instance."
                " If you need more use a Static source."
            )

        player: media.Player = media.Player()
        player.volume = volume
        player.position = (
            pan,
            0.0,
            math.sqrt(1 - math.pow(pan, 2)),
        )  # used to mimic panning with 3D audio

        # Note that the underlying attribute is pitch but "speed" is used
        # because it describes the behavior better (see #1198)
        player.pitch = speed

        player.loop = loop
        player.queue(self.source)
        player.play()
        media.Source._players.append(player)

        def _on_player_eos():
            # Some race condition within Pyglet can cause the player to be removed
            # from this list before we get to it, so we try and catch the ValueError
            # raised by the removal if it's already been removed.
            try:
                media.Source._players.remove(player)
            except ValueError:
                pass
            # There is a closure on player. To get the refcount to 0,
            # we need to delete this function.
            player.on_player_eos = None  # type: ignore  # pending https://github.com/pyglet/pyglet/issues/845

        player.on_player_eos = _on_player_eos  # type: ignore
        return player

    def stop(self, player: media.Player) -> None:
        """Stop and :py:meth:`~pyglet.media.player.Player.delete` ``player``.

        All references to it in the internal table for
        :py:class:`pyglet.media.Source` will be deleted.

        Args:
            player: A pyglet |pyglet Player| from :func:`play_sound`
                or :py:meth:`Sound.play`.
        """
        player.pause()
        player.delete()
        if player in media.Source._players:
            media.Source._players.remove(player)

    def get_length(self) -> float:
        """Get length of the loaded audio in seconds"""
        # We validate that duration is known when loading the source
        return self.source.duration  # type: ignore

    def is_complete(self, player: media.Player) -> bool:
        """``True`` if the sound is done playing."""
        # We validate that duration is known when loading the source
        return player.time >= self.source.duration  # type: ignore

    def is_playing(self, player: media.Player) -> bool:
        """``True`` if ``player`` is currently playing, otherwise ``False``.

        Args:
            player: A |pyglet Player| from :func:`play_sound` or
                :py:meth:`Sound.play`.

        Returns:
            ``True`` if the passed pyglet player is playing.
        """
        return player.playing

    def get_volume(self, player: media.Player) -> float:
        """Get the current volume.

        Args:
            player: A |pyglet Player| from :func:`play_sound` or
                :py:meth:`Sound.play`.
        Returns:
            A volume between ``0.0`` (silent) and ``1.0`` (full volume).
        """
        return player.volume  # type: ignore  # pending https://github.com/pyglet/pyglet/issues/847

    def set_volume(self, volume: float, player: media.Player) -> None:
        """Set the volume of a sound as it is playing.

        Args:
            volume: Floating point volume. 0 is silent, 1 is full.
            player: A |pyglet Player| from :func:`play_sound` or
                :py:meth:`Sound.play`.
        """
        player.volume = volume

    def get_stream_position(self, player: media.Player) -> float:
        """Return where we are in the stream. This will reset back to
        zero when it is done playing.

        Args:
            player: A |pyglet Player| from :func:`play_sound` or
                 :py:meth:`Sound.play`.
        """
        return player.time


def load_sound(path: str | Path, streaming: bool = False) -> Sound:
    """Load a file as a :py:class:`Sound` data object.

    .. important:: A :py:class:`Sound` with ``streaming=True`` loses features!

                   Neither ``loop`` nor simultaneous playbacks will work. See
                   :py;class:`Sound` and :ref:`sound-loading-modes`.

    Args:
        path: a path which may be prefixed with a
            :ref:`resource_handle <resource_handles>`.
        streaming: Boolean for determining if we stream the sound or
            load it all into memory. Set to ``True`` for long sounds to
            save memory, ``False`` for short sounds to speed playback.

    Returns:
        A :ref:playable <sound-basics-playing>` instance of a
        :py:class:`Sound` object.
    """
    # Initialize the audio driver if it hasn't been already.
    # This call is to avoid audio driver initialization
    # the first time a sound is played.
    # This call is inexpensive if the driver is already initialized.
    media.get_audio_driver()

    file_name = str(path)
    try:
        return Sound(file_name, streaming)
    except Exception as ex:
        raise FileNotFoundError(
            f'Unable to load sound file: "{file_name}". Exception: {ex}'
        ) from ex


def play_sound(
    sound: Sound | None,
    volume: float = 1.0,
    pan: float = 0.0,
    loop: bool = False,
    speed: float = 1.0,
) -> media.Player | None:
    """Try to play the ``sound`` and return a |pyglet Player|.

    The ``sound`` must be a loaded :py:class:`Sound` object. If you
    pass a path or :py:class:`str`, the function will raise a
    :py:class:`TypeError.`

    .. important:: A :py:class:`Sound` with ``streaming=True`` loses features!

                   Neither ``loop`` nor simultaneous playbacks will work. See
                   :py;class:`Sound` and :ref:`sound-loading-modes`.

    The output and return value depend on whether playback succeeded:
    .. # Note: substitutions don't really work inside tables, so the
    .. # pyglet player below is left as a normal class cross-reference.

    .. list-table::
       :header-rows: 1

       * - Success?
         - Console output
         - Return value

       * - No / ``sound`` is ``None``
         - Log a warning
         - ``None``

       * - Yes
         - N/A
         - A pyglet :py:class:`~pyglet.media.player.Player`

    To learn more about the ``streaming`` keyword and restrictions, please see:

    * :py:class:`Sound`
    * :ref:`sound-intermediate-playback-change-aspects-ongoing`
    * :ref:`sound-intermediate-playback-change-aspects-new`

    Args:
        sound: A :py:class:`Sound` instance or ``None``.
        volume: From ``0.0`` (silent) to ``1.0`` (max volume).
        pan: The left / right ear balance (``-1`` is left, ``0`` is center,
        and ``1`` is right)
        loop: ``True`` makes playback restart each time it reaches the end.
        speed: How fast to play. Slower than ``1.0`` deepens sound while
            values higher than ``1.0`` raise the pitch.

    Returns:
        A |pyglet Player| instance for this playback or
        ``None`` if playback failed.
    """
    if sound is None:
        logger.warning("Unable to play sound, no data passed in.")
        return None

    elif not isinstance(sound, Sound):
        raise TypeError(
            f"Error, got {sound!r} instead of an arcade.Sound."
            if not isinstance(sound, str | Path | bytes)
            else " Make sure to use load_sound first, then play the result with play_sound."
        )

    try:
        return sound.play(volume, pan, loop, speed)
    except Exception as ex:
        logger.warning("Error playing sound.", ex)
        return None


def stop_sound(player: media.Player) -> None:
    """Stop and delete a |pyglet Player| which is currently playing.

    Args:
        player: A pyglet |pyglet Player| from :py:func:`play_sound`
            or :py:meth:`Sound.play`.
    """

    if not isinstance(player, media.Player):
        raise TypeError(
            "stop_sound takes a media player object returned from the play_sound() command, not a "
            "loaded Sound object."
            if isinstance(player, Sound)
            else f"{player!r}"
        )

    player.pause()
    player.delete()
    if player in media.Source._players:
        media.Source._players.remove(player)



"""
An experiment trying to bug out the geometry shader sprite culling.
If the culling algorithm is wrong sprites can disappear before they
leave the screen.

Simply run the program and move draw the sprites around using the mouse.
"""

import PIL
from pyglet.math import Mat4

import arcade
from arcade.sprite import Sprite


class GeoCullingTest(arcade.Window):
    def __init__(self):
        super().__init__(1280, 720, "Cull test", resizable=True)
        self.proj = 0, self.width, 0, self.height
        self.texture = arcade.Texture(
            PIL.Image.new("RGBA", (2048, 2), (255, 255, 255, 255)),
            hash="weird_texture",
        )

        self.spritelist = arcade.SpriteList()
        self.spritelist.append(
            Sprite(
                ":resources:images/tiles/boxCrate_double.png",
                center_x=400,
                center_y=300,
                scale=6,
            )
        )
        for i in range(0, 360, 36):
            self.spritelist.append(arcade.Sprite(self.texture, center_x=400, center_y=300, angle=i))

        self.spritelist.append(
            Sprite(":resources:images/items/gold_1.png", center_x=400, center_y=300)
        )

    def on_draw(self):
        self.clear()
        self.ctx.projection_matrix = Mat4.orthogonal_projection(*self.proj, -100, 100)
        self.spritelist.draw(pixelated=True)

    def on_resize(self, width, height):
        super().on_resize(width, height)
        self.proj = 0, width, 0, height

    def on_mouse_drag(self, x: float, y: float, dx: float, dy: float, buttons: int, modifiers: int):
        self.proj = (
            self.proj[0] - dx,
            self.proj[1] - dx,
            self.proj[2] - dy,
            self.proj[3] - dy,
        )


window = GeoCullingTest()
arcade.run()



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from arcade.texture import ImageData


class ImageDataCache:
    """
    Simple cache for ImageData objects.

    These are usually cached using the absolute path to the file
    or some custom unique name if generated from code.

    We are not caching by image hash because this defeats the purpose
    of this cache. We want to obtain and image based on a value that
    can be constructed without knowing the pixel data.

    The reasoning for caching the ImageData object instead of the
    PIL.Image object is to avoid re-calculating the hash in addition
    to eliminating the need to load it and convert the pixel format.
    """

    def __init__(self):
        self._entries: dict[str, ImageData] = {}

    def put(self, name: str, image: ImageData):
        """
        Add an image to the cache.

        An entry can only be cached as either strong or weak, not both.
        If and existing entry is found, it will be replaced.

        Args:
            name: Name of the image
            image: ImageData object
        """
        from arcade.texture import ImageData

        if not isinstance(image, ImageData):
            raise TypeError("image must be an instance of ImageData")

        self._entries[name] = image

    def get(self, name: str) -> ImageData | None:
        """
        Attempts to retrieve an entry from the cache.

        Args:
            name: Name of the image
        Returns:
            ImageData instance or ``None`` if not found
        """
        return self._entries.get(name)

    def delete(self, name: str, raise_if_not_exist: bool = False) -> None:
        """
        Attempts to delete an entry from the cache.

        Args:
            name:
                Name of the image
            raise_if_not_exist:
                If ``True``, raises ``KeyError`` if the entry does not exist
        """
        try:
            del self._entries[name]
        except KeyError:
            if raise_if_not_exist:
                raise

    def flush(self):
        """Clears the cache."""
        self._entries.clear()

    def __len__(self):
        return len(self._entries)

    def __getitem__(self, name: str) -> ImageData | None:
        return self.get(name)

    def __setitem__(self, name: str, image: ImageData):
        self.put(name, image)

    def __delitem__(self, name: str):
        self.delete(name, raise_if_not_exist=True)



"""
Section Example 1:

In this Section example we divide the screen in two sections and let the user
pick a box depending on the selected Section

Note:
    - How View know nothing of what's happening inside the sections.
      Each section knows what to do.
    - Each event mouse input is handled by each Section even if the class
      it's the same (ScreenPart).
    - How on_mouse_enter/leave triggers in each Section when the mouse
      enter or leaves the section boundaries

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sections_demo_1
"""

import arcade
from arcade import SectionManager


class Box(arcade.SpriteSolidColor):
    """This is a Solid Sprite that represents a GREEN Box on the screen"""

    def __init__(self, section):
        super().__init__(100, 100, color=arcade.color.APPLE_GREEN)
        self.section = section

    def on_update(self, delta_time: float = 1 / 60):
        # update the box (this actually moves the Box by changing its position)
        self.update()

        # if we hit the ground then lay on the ground and stop movement
        if self.bottom <= 0:
            self.bottom = 0
            self.stop()

    def release(self):
        self.section.hold_box = None
        self.change_y = -10


class ScreenPart(arcade.Section):
    """
    This represents a part of the View defined by its
    boundaries (left, bottom, etc.)
    """

    def __init__(self, left: int, bottom: int, width: int, height: int, **kwargs):
        super().__init__(left, bottom, width, height, **kwargs)

        self.selected: bool = False  # if this section is selected

        self.box: Box = Box(self)  # the section Box Sprite

        # position the Box inside this section using self.left + self.width
        self.box.position = self.left + (self.width / 2), 50

        # variable that will hold the Box when it's being dragged
        self.hold_box: Box | None = None

    def on_update(self, delta_time: float):
        # call on_update on the owned Box
        self.box.on_update(delta_time)

    def on_draw(self):
        """Draw this section"""
        if self.selected:
            # Section is selected when mouse is within its boundaries
            arcade.draw_lrbt_rectangle_filled(
                self.left, self.right, self.bottom, self.top, arcade.color.GRAY
            )
            arcade.draw_text(
                f"You're are on the {self.name}",
                self.left + 30,
                self.top - 50,
                arcade.color.BLACK,
                16,
            )

        # draw the box
        arcade.draw_sprite(self.box)

    def on_mouse_drag(
        self, x: float, y: float, dx: float, dy: float, _buttons: int, _modifiers: int
    ):
        # if we hold a box, then whe move it at the same rate the mouse moves
        if self.hold_box:
            self.hold_box.position = x, y

    def on_mouse_press(self, x: float, y: float, button: int, modifiers: int):
        # if we pick a Box with the mouse, "hold" it and stop its movement
        if self.box.collides_with_point((x, y)):
            self.hold_box = self.box
            self.hold_box.stop()

    def on_mouse_release(self, x: float, y: float, button: int, modifiers: int):
        # if hold_box is True because we pick it with on_mouse_press
        # then release the Box
        if self.hold_box:
            self.hold_box.release()

    def on_mouse_enter(self, x: float, y: float):
        # select this section
        self.selected = True

    def on_mouse_leave(self, x: float, y: float):
        # unselect this section
        self.selected = False

        # if we are holding this section box and we leave the section
        # we release the box as if we release the mouse button
        if self.hold_box:
            self.hold_box.release()


class GameView(arcade.View):
    def __init__(self):
        super().__init__()

        # add sections to the view
        self.section_manager = SectionManager(self)

        # 1) First section holds half of the screen
        self.section_manager.add_section(
            ScreenPart(0, 0, self.window.width / 2, self.window.height, name="Left")
        )

        # 2) Second section holds the other half of the screen
        self.section_manager.add_section(
            ScreenPart(
                self.window.width / 2, 0, self.window.width / 2, self.window.height, name="Right"
            )
        )

    def on_show_view(self) -> None:
        self.section_manager.enable()

    def on_hide_view(self) -> None:
        self.section_manager.disable()

    def on_draw(self):
        # clear the screen
        self.clear(color=arcade.color.BEAU_BLUE)

        # draw a line separating each Section
        arcade.draw_line(
            self.window.width / 2,
            0,
            self.window.width / 2,
            self.window.height,
            arcade.color.BLACK,
            1,
        )


def main():
    # create the window
    window = arcade.Window()

    # create the custom View. Sections are initialized inside the GameView init
    view = GameView()

    # show the view
    window.show_view(view)

    # run arcade loop
    window.run()


if __name__ == "__main__":
    main()



"""
Load a Tiled map file with Levels

Artwork from: https://kenney.nl
Tiled available from: https://www.mapeditor.org/

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_tiled_map_with_levels
"""
import arcade

TILE_SPRITE_SCALING = 0.5
PLAYER_SCALING = 0.6

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Tiled Map with Levels Example"
SPRITE_PIXEL_SIZE = 128
GRID_PIXEL_SIZE = SPRITE_PIXEL_SIZE * TILE_SPRITE_SCALING
CAMERA_PAN_SPEED = 0.15

# Physics
MOVEMENT_SPEED = 5
JUMP_SPEED = 23
GRAVITY = 1.1


class GameView(arcade.View):
    """Main application class."""

    def __init__(self):
        """
        Initializer
        """
        super().__init__()

        # Tilemap Object
        self.tile_map = None

        # Sprite lists
        self.player_list = None

        # Set up the player
        self.score = 0
        self.player_sprite = None

        self.physics_engine = None
        self.end_of_map = 0
        self.game_over = False

        self.game_camera = None
        self.gui_camera = None
        self.camera_bounds = None

        self.fps_text = None
        self.game_over_text = None

        self.level = 1
        self.max_level = 2

    def setup(self):
        """Set up the game and initialize the variables."""

        # Sprite lists
        self.player_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=PLAYER_SCALING,
        )

        self.game_camera = arcade.Camera2D()
        self.gui_camera = arcade.Camera2D()

        self.fps_text = arcade.Text('FPS:', 10, 10, arcade.color.BLACK, 14)
        self.game_over_text = arcade.Text(
            'GAME OVER',
            self.window.center_x,
            self.window.center_y,
            arcade.color.BLACK, 30,
            anchor_x='center',
            anchor_y='center'
        )

        # Starting position of the player
        self.player_sprite.center_x = 128
        self.player_sprite.center_y = 64
        self.player_list.append(self.player_sprite)

        self.load_level(self.level)

        self.game_over = False

    def load_level(self, level):
        # layer_options = {"Platforms": {"use_spatial_hash": True}}

        # Read in the tiled map
        self.tile_map = arcade.load_tilemap(
            f":resources:tiled_maps/level_{level}.json", scaling=TILE_SPRITE_SCALING
        )

        # --- Walls ---

        # Calculate the right edge of the my_map in pixels
        self.end_of_map = self.tile_map.width * GRID_PIXEL_SIZE

        self.physics_engine = arcade.PhysicsEnginePlatformer(
            self.player_sprite,
            self.tile_map.sprite_lists["Platforms"],
            gravity_constant=GRAVITY,
        )

        # --- Other stuff
        # Set the background color
        if self.tile_map.background_color:
            self.window.background_color = self.tile_map.background_color

        max_x = GRID_PIXEL_SIZE * self.tile_map.width
        max_y = GRID_PIXEL_SIZE * self.tile_map.height
        limit_y = max_y > self.window.height
        self.camera_bounds = arcade.LRBT(
            self.window.width / 2.0,
            max_x - self.window.width / 2.0,
            self.window.height / 2.0,
            max_y - (self.window.height / 2.0 if limit_y else 0.0)
        )

        # Reset cam
        self.game_camera.position = self.window.center

    def on_draw(self):
        """
        Render the screen.
        """
        # This command has to happen before we start drawing
        self.clear()

        with self.game_camera.activate():
            # Draw all the sprites.
            self.tile_map.sprite_lists["Platforms"].draw()
            self.player_list.draw()

        with self.gui_camera.activate():
            # Put the text on the screen.
            # Adjust the text position based on the view port so that we don't
            # scroll the text too.
            output = f"FPS: {1/self.window.delta_time:.0f}"
            arcade.draw_text(
                output, 10, 20, arcade.color.BLACK, 14
            )

            if self.game_over:
                self.game_over_text.position = self.window.center
                self.game_over_text.draw()

    def on_key_press(self, key, modifiers):
        """
        Called whenever the mouse moves.
        """
        if key == arcade.key.UP:
            if self.physics_engine.can_jump():
                self.player_sprite.change_y = JUMP_SPEED
        elif key == arcade.key.LEFT:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_x = MOVEMENT_SPEED

    def on_key_release(self, key, modifiers):
        """
        Called when the user presses a mouse button.
        """
        if key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_x = 0

    def on_update(self, delta_time):
        """Movement and game logic"""

        if self.player_sprite.right >= self.end_of_map:
            if self.level < self.max_level:
                self.level += 1
                self.load_level(self.level)
                self.player_sprite.center_x = 128
                self.player_sprite.center_y = 64
                self.player_sprite.change_x = 0
                self.player_sprite.change_y = 0
            else:
                self.game_over = True

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        if not self.game_over:
            self.physics_engine.update()

            # --- Manage Scrolling ---
            self.game_camera.position = arcade.math.smerp_2d(
                self.game_camera.position,
                self.player_sprite.position,
                delta_time,
                CAMERA_PAN_SPEED
            )
            self.game_camera.position = arcade.camera.grips.constrain_xy(
                self.game_camera.view_data,
                self.camera_bounds
            )


def main():
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    game = GameView()
    game.setup()

    window.show_view(game)
    window.run()


if __name__ == "__main__":
    main()
